from __future__ import annotations

import contextlib
import dataclasses
import inspect
import json
import os
import re
from datetime import timedelta
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import yaml
from typing_extensions import ParamSpec, TypeAlias
from yaml.scanner import ScannerError

from chalk import Environments, OfflineResolver, OnlineResolver, Tags
from chalk._lsp.error_builder import SQLFileResolverErrorBuilder
from chalk.features import DataFrame, Feature, FeatureNotFoundException, Features
from chalk.features.feature_set import CURRENT_FEATURE_REGISTRY
from chalk.features.namespace_context import build_namespaced_name
from chalk.features.namespace_context import namespace as namespace_ctx
from chalk.features.pseudofeatures import Now
from chalk.features.resolver import Cron, ResolverArgErrorHandler, StreamResolver
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
from chalk.sql._internal.integrations.cloudsql import CloudSQLSourceImpl
from chalk.sql._internal.integrations.mysql import MySQLSourceImpl
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
from chalk.sql._internal.integrations.snowflake import SnowflakeSourceImpl
from chalk.sql._internal.integrations.spanner import SpannerSourceImpl
from chalk.sql._internal.integrations.sqlite import SQLiteSourceImpl
from chalk.sql._internal.sql_settings import SQLResolverSettings
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.sql.finalized_query import Finalizer
from chalk.streams import KafkaSource, get_resolver_error_builder
from chalk.streams.base import StreamSource
from chalk.streams.types import StreamResolverSignature
from chalk.utils import MachineType, notebook
from chalk.utils.collections import get_unique_item, get_unique_item_if_exists
from chalk.utils.duration import CronTab, Duration, parse_chalk_duration, timedelta_to_duration
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.string import to_snake_case

P = ParamSpec("P")
T = TypeVar("T")

FeatureReference: TypeAlias = Union[str, Any]

if TYPE_CHECKING:
    import sqlglot.expressions
    from pydantic import BaseModel, ValidationError

    from chalk.sql import BaseSQLSourceProtocol, SQLSourceGroup

    def validator(*args: str, pre: bool = ...) -> Callable[[Callable[P, T]], Callable[P, T]]:
        ...

else:
    try:
        from pydantic.v1 import BaseModel, ValidationError, validator
    except ImportError:
        from pydantic import BaseModel, ValidationError, validator

_SOURCES: Mapping[str, Union[Type[BaseSQLSource], Type[StreamSource]]] = {
    "snowflake": SnowflakeSourceImpl,
    "postgres": PostgreSQLSourceImpl,
    "postgresql": PostgreSQLSourceImpl,
    "mysql": MySQLSourceImpl,
    "bigquery": BigQuerySourceImpl,
    "cloudsql": CloudSQLSourceImpl,
    "redshift": RedshiftSourceImpl,
    "sqlite": SQLiteSourceImpl,
    "kafka": KafkaSource,
    "spanner": SpannerSourceImpl,
}

_SQLGLOT_DIALECTS = frozenset(
    (
        "snowflake",
        "postgres",
        "mysql",
        "bigquery",
        "redshift",
        "sqlite",
        "databricks",
    )
)
"""These dialects are used if a "kind" of source is listed, rather than a specific source."""

_RESOLVER_TYPES = {
    "offline": OfflineResolver,
    "batch": OfflineResolver,
    "online": OnlineResolver,
    "realtime": OnlineResolver,
    "stream": StreamResolver,
    "streaming": StreamResolver,
}

CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX = ".chalk.sql"


class IncrementalSettingsSQLFileResolver(BaseModel):
    incremental_column: Optional[str]

    lookback_period: Optional[Duration]

    mode: Literal["row", "group", "parameter"] = "row"

    incremental_timestamp: Optional[Literal["feature_time", "resolver_execution_time"]] = "feature_time"

    @validator("lookback_period")
    @classmethod
    def validate_lookback_period(cls, value: Optional[str]):
        if value is None:
            return None
        if isinstance(value, timedelta):
            return value
        parse_chalk_duration(value)
        return value

    @validator("mode")
    @classmethod
    def validate_mode(cls, mode: Literal["row", "group", "parameter"], values: Dict[str, Any]):
        if mode in ["row", "group"] and not values["incremental_column"]:
            raise ValueError("'incremental_column' must be set if mode is 'row' or 'group'.")
        return mode


class CommentDict(BaseModel):
    total: Optional[bool]
    source: Optional[str]
    resolves: Optional[str]
    namespace: Optional[str]
    incremental: Optional[IncrementalSettingsSQLFileResolver]
    tags: Optional[List[str]]
    environment: Optional[List[str]]
    count: Optional[Literal[1, "one", "one_or_none", "all"]]
    cron: Optional[Any]
    machine_type: Optional[str]
    owner: Optional[str]
    type: Optional[str]
    timeout: Optional[str]
    fields: Optional[Dict[str, str]]
    unique_on: Optional[List[str]]
    partitioned_by: Optional[List[str]]

    @validator("tags", "environment", "unique_on", "partitioned_by", pre=True)
    @classmethod
    def validate_list_inputs(cls, value: Union[str, List[str], None]):
        if value is None:
            return None
        if isinstance(value, list):
            return value
        if isinstance(value, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            return [value]
        raise ValueError(f"Value {value} must be a string or a list of strings.")

    @validator("cron", pre=True)
    @classmethod
    def validate_cron_input(cls, value: Any):
        if value is None:
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, timedelta):
            return value
        if isinstance(value, Cron):
            return value
        raise ValueError(f"Value {value} must be a string or a Cron")

    @validator("timeout")
    @classmethod
    def validate_timedelta(cls, string: Optional[str]):
        if string is None:
            return None
        parse_chalk_duration(string)
        return string

    @validator("type")
    @classmethod
    def validate_type(cls, resolver_type: str | None):
        if resolver_type is None:
            return None
        if resolver_type not in _RESOLVER_TYPES:
            raise ValueError(
                (
                    f"Resolver type '{resolver_type}' not supported. "
                    f"'online', 'offline' and 'streaming' are supported options"
                )
            )
        return resolver_type


@dataclasses.dataclass(frozen=True)
class ResolverError:
    """Generic class for returning errors at any point during resolution process"""

    display: str
    path: str
    parameter: Optional[str]


@dataclasses.dataclass(frozen=True)
class ResolverResult:
    """Chief return class with resolver we actually use"""

    resolver: Optional[Union[OnlineResolver, OfflineResolver, StreamResolver]]
    errors: List[ResolverError]
    db: Optional[Union[BaseSQLSource, StreamSource, SQLSourceGroup]]
    fields: Optional[Dict[str, str]]
    args: Optional[Dict[str, str]]
    data_lineage: Optional[Dict[str, Dict[str, Dict[str, List[str]]]]] = None


@dataclasses.dataclass(frozen=True)
class SQLStringResult:
    """Class for getting the sql string from the file"""

    path: str
    sql_string: Optional[str]
    error: Optional[ResolverError]
    override_comment_dict: Optional[CommentDict] = None
    override_name: Optional[str] = None
    autogenerated: bool = False

    def __post_init__(self):
        # Validation: if autogenerated is True, override_name must not be None
        if self.autogenerated and self.override_name is None:
            raise ValueError("override_name must be non-None if autogenerated is True.")

    @classmethod
    def fail(cls, display_error: str, path: str) -> "SQLStringResult":
        return cls(
            path=path,
            sql_string=None,
            error=ResolverError(display=display_error, path=path, parameter=None),
        )


@dataclasses.dataclass(frozen=True)
class GlotResult:
    """Class for editing the sql string, and using sqlglot on sql string"""

    sql_string: str
    glot: Optional[Union[sqlglot.expressions.Select, sqlglot.expressions.Union]]
    args: Dict[str, str]
    default_args: List[Union[Optional[str], ellipsis]]
    comment_dict: Optional[CommentDict]
    docstring: Optional[str]
    errors: List[ResolverError]
    source: Union[BaseSQLSource, StreamSource, SQLSourceGroup, None]


@dataclasses.dataclass(frozen=True)
class ParseResult:
    """Class for important info gathered from glot"""

    sql_string: str
    comment_dict: CommentDict
    fields: Dict[str, str]
    namespace: str
    source: Union[BaseSQLSource, StreamSource, SQLSourceGroup, None]
    docstring: Optional[str]
    errors: List[ResolverError]

    # data_lineage is stored as a map { data_source: { table: { column: feature } } }
    data_lineage: Optional[Dict[str, Dict[str, Dict[str, List[str]]]]] = None


_filepath_to_sql_string: dict[str, str] = {}
"""Mapping from filepath to sql string. Used to skip reimporting the same sql file if the content is identical to what was already imported
If the content is different, then we'll import it again, but we may error later when attempting to add the resolver to the registry if we don't
allow overrides and we're not in a notebook.
"""


def get_sql_file_resolvers(
    *,
    sql_file_resolve_location: Path,
    sources: Sequence[BaseSQLSource | SQLSourceGroup],
    has_import_errors: bool,
) -> Iterable[ResolverResult]:
    """Iterate through all `.chalk.sql` filepaths, gather the sql strings, and get a resolver hopefully for each."""
    for dp, dn, fn in os.walk(os.path.expanduser(sql_file_resolve_location)):
        del dn  # unused
        for f in fn:
            filepath = os.path.join(dp, f)
            if not filepath.endswith(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX):
                continue
                # Already imported this file. Skipping it, assuming it did not change since the last time we imported it
            sql_string_result = _get_sql_string(filepath)
            existing_sql_string = _filepath_to_sql_string.get(filepath)
            if sql_string_result.sql_string is not None:
                if existing_sql_string is not None and existing_sql_string == sql_string_result.sql_string:
                    # The sql file is identical to what was already imported. skip it
                    continue
                _filepath_to_sql_string[filepath] = sql_string_result.sql_string
            yield get_sql_file_resolver(sources, sql_string_result, has_import_errors)
    for sql_string_result in _GENERATED_SQL_FILE_RESOLVER_REGISTRY.get_generated_sql_file_resolvers():
        yield get_sql_file_resolver(sources, sql_string_result, has_import_errors)


def get_sql_file_resolvers_from_paths(
    *,
    sources: Sequence[BaseSQLSource | SQLSourceGroup],
    paths: List[str],
    has_import_errors: bool,
) -> Iterable[ResolverResult]:
    for p in paths:
        sql_string_result = _get_sql_string(path=p)
        existing_sql_string = _filepath_to_sql_string.get(p)
        if sql_string_result.sql_string is not None:
            if existing_sql_string is not None and existing_sql_string == sql_string_result.sql_string:
                # The sql file is identical to what was already imported. skip it
                continue
            _filepath_to_sql_string[p] = sql_string_result.sql_string
        yield get_sql_file_resolver(
            sources=sources,
            sql_string_result=sql_string_result,
            has_import_errors=has_import_errors,
        )
    for sql_string_result in _GENERATED_SQL_FILE_RESOLVER_REGISTRY.get_generated_sql_file_resolvers():
        yield get_sql_file_resolver(sources, sql_string_result, has_import_errors)


def get_sql_file_resolver(
    sources: Iterable[BaseSQLSource | SQLSourceGroup],
    sql_string_result: SQLStringResult,
    has_import_errors: bool = False,
) -> ResolverResult:
    from chalk.sql import SQLSourceGroup

    registry_features = CURRENT_FEATURE_REGISTRY.get().get_feature_sets()

    assert sql_string_result.sql_string is not None, f"SQL string from path {sql_string_result.path} should not be None"
    error_builder = SQLFileResolverErrorBuilder(
        uri=sql_string_result.path, sql_string=sql_string_result.sql_string, has_import_errors=has_import_errors
    )

    """Parse the sql strings and get a ResolverResult from each"""
    if sql_string_result.error:
        return ResolverResult(
            resolver=None,
            errors=[sql_string_result.error],
            db=None,
            fields=None,
            args=None,
        )
    path = sql_string_result.path

    errors: List[ResolverError] = []
    glot_result = _get_sql_glot(
        sql_string=sql_string_result.sql_string,
        path=path,
        sources=sources,
        error_builder=error_builder,
        override_comment_dict=sql_string_result.override_comment_dict,
    )
    if glot_result.errors:
        return ResolverResult(
            resolver=None,
            errors=glot_result.errors,
            db=None,
            fields=None,
            args=None,
        )

    parsed = _parse_glot(
        glot_result=glot_result,
        path=path,
        error_builder=error_builder,
    )
    if parsed.errors:
        return ResolverResult(
            resolver=None,
            errors=parsed.errors,
            db=parsed.source,
            fields=None,
            args=None,
        )
    with (
        namespace_ctx(parsed.comment_dict.namespace)
        if parsed.comment_dict and parsed.comment_dict.namespace
        else contextlib.nullcontext()
    ):
        # validate inputs and outputs as real features in graph
        inputs: List[Feature] = []
        for arg in glot_result.args.values():
            try:
                inputs.append(Feature.from_root_fqn(build_namespaced_name(name=arg)))
            except FeatureNotFoundException:  # other exceptions will be caught eventually
                message = f"SQL file resolver references an input feature '{arg}' which does not exist."
                if arg in registry_features:
                    message += f" It appears '{arg}' is a feature class, not a feature."
                try:
                    error_builder.add_diagnostic_with_spellcheck(
                        spellcheck_item=arg,
                        spellcheck_candidates=[
                            feature.fqn
                            for feature in registry_features[build_namespaced_name(name=parsed.namespace)].features
                        ],
                        message=message,
                        code="152",
                        label="input feature not recognized",
                        range=error_builder.variable_range_by_name(arg),
                    )
                except Exception:
                    # it's possible the graph is incomplete due to other import errors and we cannot find spellcheck candidates
                    # in this case, a simple error will do
                    error_builder.add_diagnostic(
                        message=message,
                        code="152",
                        label="input feature not recognized",
                        range=error_builder.variable_range_by_name(arg),
                    )
                errors.append(
                    ResolverError(
                        display=message,
                        path=path,
                        parameter=arg,
                    )
                )
        outputs: List[Feature] = []
        query_fields: Dict[str, str] = {}
        for variable, output in parsed.fields.items():
            message = f"SQL file resolver references an output feature '{output}' which does not exist. "
            if output.endswith("*"):
                features = registry_features[build_namespaced_name(name=parsed.namespace)].features
                for feature in features:
                    if (
                        feature.is_scalar
                        and not feature.is_autogenerated
                        and not feature.is_pseudofeature
                        and not feature.is_windowed
                    ):
                        outputs.append(feature)
                continue

            unrecognized_output = False
            try:
                feature = Feature.from_root_fqn(output)
                outputs.append(feature)
                query_fields[variable] = output
            except FeatureNotFoundException:
                # there's only one way to have a recognized output here, if the output is in the comment dict 'fields'
                unrecognized_output = True
                if parsed.comment_dict.fields is not None:
                    split = output.split(".", 1)
                    if len(split) > 1:
                        namespace = split[0]
                        output_name = split[1]
                        if output_name in parsed.comment_dict.fields:
                            field = parsed.comment_dict.fields.get(output_name)
                            output = f"{namespace}.{field}"
                            try:
                                feature = Feature.from_root_fqn(build_namespaced_name(name=output))
                                outputs.append(feature)
                                query_fields[output_name] = output
                                unrecognized_output = False
                            except:
                                pass
            except Exception as e:
                # it's possible the graph is incomplete due to other import errors and we cannot even use Feature.from_root_fqn without
                # throwing an unhandled error. In this case, we don't addi a diagnostic because we are confident the error
                # will surface later with a better stack
                errors.append(
                    ResolverError(
                        display=f"{message}: {e}",
                        path=path,
                        parameter=output,
                    )
                )
                continue

            if unrecognized_output:
                value = output.split(".", 1)[1] if "." in output else output
                assert glot_result.glot is not None, f"Failed to parse {glot_result.sql_string}"
                try:
                    error_builder.add_diagnostic_with_spellcheck(
                        spellcheck_item=output,
                        spellcheck_candidates=[
                            feature.fqn
                            for feature in registry_features[build_namespaced_name(name=parsed.namespace)].features
                        ],
                        message=message,
                        code="153",
                        label="output feature not recognized",
                        range=error_builder.value_range_by_name(glot_result.glot, value),
                    )
                except Exception:
                    # it's possible the graph is incomplete due to other import errors and we cannot find spellcheck candidates
                    # in this case, we don't add a diagnostic because we are confident the error will surface later with a better stack
                    pass
                errors.append(
                    ResolverError(
                        display=message,
                        path=path,
                        parameter=output,
                    )
                )
        if errors:
            return ResolverResult(resolver=None, errors=errors, db=parsed.source, fields=None, args=glot_result.args)

        if len(outputs) == 0:
            message = (
                "SQL file resolver has no detected outputs. Make sure that all select outputs are aliased to features "
                "defined within the feature set referenced by the `resolves` parameter."
            )
            error_builder.add_diagnostic(
                message=message,
                code="154",
                label="no outputs detected",
                range=error_builder.sql_range(),
            )
            errors.append(ResolverError(display=message, path=path, parameter=None))

        if parsed.comment_dict.unique_on is not None:
            unique_on = _validate_feature_strings_in_comments(
                feature_strings=parsed.comment_dict.unique_on,
                error_builder=error_builder,
                namespace=parsed.namespace,
                path=path,
                errors=errors,
                outputs=outputs,
                arg_name="unique_on",
            )
        else:
            unique_on = None
        if parsed.comment_dict.partitioned_by is not None:
            partitioned_by = _validate_feature_strings_in_comments(
                feature_strings=parsed.comment_dict.partitioned_by,
                error_builder=error_builder,
                namespace=parsed.namespace,
                path=path,
                errors=errors,
                outputs=outputs,
                arg_name="partitioned_by",
            )
        else:
            partitioned_by = None
        resolver_type_str = parsed.comment_dict.type if parsed.comment_dict.type else "online"
        resolver_type = _RESOLVER_TYPES[resolver_type_str]

        if resolver_type == StreamResolver:
            return _get_stream_resolver(path, glot_result, parsed, outputs, error_builder)

        incremental_dict = parsed.comment_dict.incremental.dict() if parsed.comment_dict.incremental else None
        finalizer = parse_finalizer(parsed.comment_dict.count)
        source = parsed.source

        if not isinstance(source, (BaseSQLSource, SQLSourceGroup)):
            raise ValueError(f"Datasource '{source}' is not configured. Is the driver installed?")

        # function for online resolver to process
        def fn(
            *input_values: Any,
            database: BaseSQLSource | SQLSourceGroup = source,
            sql_query: str = parsed.sql_string,
            field_dict: Dict[str, str] = query_fields,
            args_dict: Dict[str, str] = glot_result.args,
            incremental: Optional[Dict[str, Any]] = incremental_dict,
        ):
            arg_dict = {arg: input_value for input_value, arg in zip(input_values, args_dict.keys())}
            func = database.query_string(
                query=sql_query,
                fields=field_dict,
                args=arg_dict,
            )
            if incremental:
                func = func.incremental(**incremental)
            elif finalizer == Finalizer.ONE:
                func = func.one()
            elif finalizer == Finalizer.ONE_OR_NONE:
                func = func.one_or_none()
            elif finalizer == Finalizer.ALL:
                func = func.all()
            return func

        if errors:
            return ResolverResult(
                resolver=None,
                errors=errors,
                db=parsed.source,
                fields=parsed.fields,
                args=glot_result.args,
            )
        if finalizer is None:
            # If the root ns of the inputs is the same as what is being resolved, then it is always return one
            input_root_namespace = get_unique_item_if_exists(x.root_namespace for x in inputs if x.fqn != Now.fqn)
            output_root_namespace = get_unique_item(x.root_namespace for x in outputs)
            if input_root_namespace is not None and input_root_namespace == output_root_namespace:
                finalizer = Finalizer.ONE

        if finalizer == Finalizer.ONE or finalizer == Finalizer.ONE_OR_NONE:
            output = Features[tuple(outputs)]
        else:
            output = Features[DataFrame[tuple(outputs)]]
        if incremental_dict is None:
            incremental_settings = None
        else:
            incremental_mode = incremental_dict.get("mode", "row")
            if incremental_mode not in ["row", "group", "parameter"]:
                raise ValueError(f"Invalid incremental mode: {incremental_mode}")
            if incremental_dict.get("lookback_period") is not None:
                lookback_period = parse_chalk_duration(incremental_dict["lookback_period"])
            else:
                lookback_period = None
            incremental_timestamp = incremental_dict.get("incremental_timestamp", "feature_time")
            if incremental_timestamp not in ["feature_time", "resolver_execution_time"]:
                raise ValueError(f"Invalid incremental timestamp: {incremental_timestamp}")
            incremental_settings = IncrementalSettings(
                mode=incremental_mode,
                incremental_column=incremental_dict.get("incremental_column"),
                lookback_period=lookback_period,
                incremental_timestamp=incremental_timestamp,
            )

        default_args: List[Optional[ResolverArgErrorHandler]] = [
            None if default_value is ... else ResolverArgErrorHandler(default_value)
            for default_value in glot_result.default_args
        ]

        filename = os.path.basename(path)
        # attempt to instantiate the resolver
        try:
            if resolver_type not in (OnlineResolver, OfflineResolver):
                raise ValueError(f"Resolver type '{resolver_type}' is not supported for .chalk.sql resolvers")
            if not isinstance(parsed.source, (BaseSQLSource, SQLSourceGroup)):
                raise ValueError(f"Datasource '{parsed.source}' is not configured. Is the driver installed")
            if sql_string_result.autogenerated:
                """Autogenerated resolvers are expected to have a name passed in through make_sql_file_resolver"""
                assert (
                    sql_string_result.override_name is not None
                ), "override_name must be non-None if autogenerated is True."
                fqn = sql_string_result.override_name
            else:
                """Non-autogenerated resolvers will have a name derived from the filename"""
                fqn = filename.replace(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX, "")

            resolver = resolver_type(
                filename=path,
                function_definition=sql_string_result.sql_string,
                fqn=fqn,
                doc=parsed.docstring,
                inputs=inputs,
                output=output,
                fn=fn,
                environment=parsed.comment_dict.environment,
                tags=parsed.comment_dict.tags,
                cron=parsed.comment_dict.cron,
                machine_type=parsed.comment_dict.machine_type,
                when=None,
                state=None,
                default_args=default_args,
                owner=parsed.comment_dict.owner,
                timeout=parsed.comment_dict.timeout,
                is_sql_file_resolver=True,
                data_sources=[parsed.source],
                source_line=None,
                lsp_builder=get_resolver_error_builder(fn),
                static=False,
                parse=None,
                resource_hint=None,
                total=True if parsed.comment_dict.total else False,
                autogenerated=sql_string_result.autogenerated,
                unique_on=unique_on,
                partitioned_by=partitioned_by,
                data_lineage=parsed.data_lineage,
                sql_settings=SQLResolverSettings(
                    finalizer=finalizer if finalizer is not None else Finalizer.ALL,
                    fields_root_fqn=query_fields,
                    incremental_settings=incremental_settings,
                    params_to_root_fqn=glot_result.args,
                ),
            )
        except Exception as e:
            raise e
            message = f"SQL file resolver '{filename}'  could not be instantiated, {e}"
            error_builder.add_diagnostic(
                message=message,
                code="155",
                label="resolver instantiation failed",
                range=error_builder.full_range(),
            )
            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=None,
                )
            )
            return ResolverResult(
                resolver=None, errors=errors, db=parsed.source, fields=parsed.fields, args=glot_result.args
            )

        return ResolverResult(
            resolver=resolver,
            errors=errors,
            db=parsed.source,
            fields=parsed.fields,
            args=glot_result.args,
            data_lineage=parsed.data_lineage,
        )


def _get_sql_string(path: str) -> SQLStringResult:
    """Attempt to get a sql string from a filepath and gracefully exit if unable to"""

    if not path.endswith(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX):
        return SQLStringResult.fail(display_error=f"sql resolver file '{path}' must end in '.chalk.sql'", path=path)
    sql_string = None
    if os.path.isfile(path):
        with open(path) as f:
            sql_string = f.read()
    else:
        frame = inspect.currentframe()
        assert frame is not None, "could not inspect current frame"
        caller_frame = frame.f_back
        del frame
        assert caller_frame is not None, "could not inspect caller frame"
        caller_filename = inspect.getsourcefile(caller_frame)
        assert caller_filename is not None, "could not find caller filename"
        dir_path = os.path.dirname(os.path.realpath(caller_filename))
        if isinstance(path, bytes):
            path = path.decode("utf-8")
        relative_path = os.path.join(dir_path, path)
        if os.path.isfile(relative_path):
            with open(relative_path) as f:
                sql_string = f.read()
    if sql_string is None:
        return SQLStringResult.fail(display_error=f"Cannot find file '{path}'", path=path)
    return SQLStringResult(path=path, sql_string=sql_string, error=None)


def _get_data_lineage(sql: str) -> Dict[str, Dict[str, List[str]]]:
    try:
        import sqlglot
        from sqlglot import exp
        from sqlglot.optimizer.scope import build_scope
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    try:
        # Parse the SQL into an abstract syntax tree (AST)
        ast = sqlglot.parse_one(sql)
        # TODO: @melrchen parse CTE's correctly
        for expression in ast.expressions:
            if isinstance(expression, exp.With):
                return {}

        root = build_scope(ast)
        #  assuming one source for simple query
        if not root or len(root.sources.keys()) > 1:
            return {}
        table = list(root.sources.keys())[0]

        lineage: Dict[str, Dict[str, List[str]]] = {table: {}}

        # get feature and column names from the query
        def process_select(select_expr: sqlglot.Expression):
            for expr in select_expr.expressions:
                if isinstance(expr, exp.Alias):
                    output_column = expr.alias
                    input_columns: List[str] = extract_columns(expr.this)
                elif isinstance(expr, exp.Column):
                    output_column = expr.name
                    input_columns: List[str] = [expr.name]
                elif isinstance(expr, exp.Binary):
                    output_column = expr.sql()  # For columns like `name + size`
                    input_columns = extract_columns(expr)
                else:
                    continue  # Skip expressions that don't contribute to lineage

                lineage[table][output_column] = input_columns

        # Extract columns from expressions
        def extract_columns(expr: sqlglot.Expression) -> List[str]:
            columns = []
            if isinstance(expr, exp.Column):
                columns.append(expr.name)
            elif isinstance(expr, exp.Func):
                for col in expr.find_all(exp.Column):
                    columns.append(col.name)
            elif isinstance(expr, exp.Binary):
                columns.extend(extract_columns(expr.left))
                columns.extend(extract_columns(expr.right))
            return columns

        # Process the main SELECT expression
        for select_expr in ast.find_all(exp.Select):
            process_select(select_expr)

        return lineage
    except Exception:
        return {}


@dataclasses.dataclass
class EscapedSqlString:
    escaped_sql_string: str
    args: dict[str, str]  # sql string -> input feature string
    default_args: List[Union[str, None, ellipsis]]
    errors: List[ResolverError]


def escape_sql_params(
    sql_string: str,
    path: str,
    error_builder: SQLFileResolverErrorBuilder,
) -> EscapedSqlString:
    """
    Chalk allows people to write SQL resolvers that reference features using ${} syntax, such as:
    ```
    SELECT id, burrito_name FROM burritos_table where id=${burrito.id}
    ```

    However this isn't valid sql -- we replace each occurrence of ${} w/ a placeholder name and
    return the escaped SQL string w/ additional info about the params & relevant Chalk features.
    :param sql_string: String of hte original SQL resolver
    :param path: For error reporting, filepath of the SQL resolver
    :param error_builder: For LSP errors
    """
    args = {}  # sql string -> input feature string

    # In order to ensure that the variables are ordered deterministically, we use a `dict`
    # instead of a `set`, since `dict` ensures iteration order matches insertion order.
    variables = {key: None for key in re.findall("\\${.*?\\}", sql_string)}
    errors: List[ResolverError] = []
    default_args: List[Union[Optional[str], ellipsis]] = []
    # replace ?{variable_name} with :variable_name for sqlalchemy, and keep track of input args necessary
    for variable_pattern in variables:
        has_default_arg = False
        variable = variable_pattern[2:-1]  # cut off ${ and }
        for split_var in ("|", " or "):  # default argument
            # TODO cannot parse something like {Transaction.category or "Waffles or Pancakes"} yet
            if split_var in variable:
                split = variable.split(split_var)
                if len(split) != 2:
                    message = (
                        f"If character '|' is used, both variable name and default value must be "
                        f"specified in '({variable})' like '?{{variable_name | \"default_value\"}}"
                    )
                    error_builder.add_diagnostic(
                        message=message,
                        code="140",
                        label="invalid variable",
                        range=error_builder.variable_range_by_name(variable),
                    )
                    errors.append(
                        ResolverError(
                            display=message,
                            path=path,
                            parameter=None,
                        )
                    )
                else:  # has default argument
                    variable = split[0].strip()
                    default_arg = split[1].strip()
                    default_arg = json.loads(default_arg)
                    f = Feature.from_root_fqn(variable)
                    default_arg = f.converter.from_json_to_rich(default_arg)
                    default_args.append(default_arg)
                    has_default_arg = True
        if not has_default_arg:
            default_args.append(...)
        period_replaced = variable.replace(".", "_")
        sql_safe_str = f"__chalk_{period_replaced}__"
        sql_string = sql_string.replace(variable_pattern, f":{sql_safe_str}")
        args[sql_safe_str] = variable

    return EscapedSqlString(escaped_sql_string=sql_string, args=args, errors=errors, default_args=default_args)


def _get_sql_glot(
    sql_string: str,
    path: str,
    sources: Iterable[BaseSQLSource | SQLSourceGroup],
    error_builder: SQLFileResolverErrorBuilder,
    override_comment_dict: Optional[CommentDict] = None,
) -> GlotResult:
    """Get sqlglot from sql string and gracefully exit if unable to"""
    try:
        import sqlglot
        import sqlglot.errors
        import sqlglot.expressions
        import sqlglot.optimizer.scope
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    errors: List[ResolverError] = []
    escaped_sql = escape_sql_params(sql_string=sql_string, path=path, error_builder=error_builder)
    errors.extend(escaped_sql.errors)
    args = escaped_sql.args
    default_args = escaped_sql.default_args
    sql_string = escaped_sql.escaped_sql_string

    comments = ""
    docstring = ""
    comment_row_to_file_row: Dict[int, int] = {}
    comment_row_offset: Dict[int, int] = {}
    comment_line_counter = 0
    """
    Comments and docstrings are required to be at the beginning of the file resolver.
    Thus, we need to make sure that for ranges, comments are mapped to the right line number
    as docstrings are not included in the comments variable above.

    We also need to keep track of the offset of the comment in order to properly detect the range.
    """
    for line_no, comment in enumerate(sql_string.splitlines()):
        if comment.strip().startswith("--"):
            stripped_comment = comment.strip().replace("--", "")
            comment_offset = len(comment) - len(stripped_comment)
            if stripped_comment.strip().startswith("-"):
                comments += f"{stripped_comment}\n"
            else:
                if count_colons(stripped_comment) != 1:
                    docstring += f"{stripped_comment.strip()}\n"
                else:
                    comments += f"{stripped_comment}\n"
                    comment_row_to_file_row[comment_line_counter] = line_no
                    comment_row_offset[comment_line_counter] = comment_offset
                    comment_line_counter += 1
        else:
            break

    if len(comments) == 0 and override_comment_dict is None:
        message = "SQL file resolvers require comments that describe key-value pairs in YAML form."
        error_builder.add_diagnostic(
            message=message,
            code="141",
            label="missing comments",
            range=error_builder.full_range(),
            code_href="https://docs.chalk.ai/docs/sql#sql-file-resolvers",
        )
        errors.append(
            ResolverError(
                display=message,
                path=path,
                parameter=None,
            )
        )
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=None,
            docstring=docstring,
            errors=errors,
            source=None,
        )
    if len(comments) != 0:
        try:
            comment_dict: Dict[str, Any] = yaml.safe_load(comments)
        except Exception as e:
            message = f"SQL File resolver comments must have key-values in YAML form: {e}"
            if isinstance(e, ScannerError) and e.problem_mark is not None:
                comment_line_no = e.problem_mark.line
                comment_col_no = e.problem_mark.column
                file_line_no = comment_row_to_file_row[comment_line_no]
                file_comment_no = comment_row_offset[comment_line_no] + comment_col_no
                error_builder.add_diagnostic(
                    message=message,
                    code="142",
                    label="Could not parse comments as YAML",
                    range=error_builder.custom_range(line_no=file_line_no + 1, col=file_comment_no + 1),
                )

            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=comments,
                )
            )
            return GlotResult(
                sql_string=sql_string,
                glot=None,
                args=args,
                default_args=default_args,
                comment_dict=None,
                docstring=docstring,
                errors=errors,
                source=None,
            )
    else:
        comment_dict = {}
    if override_comment_dict is not None:
        for key, value in override_comment_dict.dict().items():
            if value is not None:
                comment_dict[key] = value
    if "source" not in comment_dict:
        """This would be caught by the following BaseModel.parse_obj() but the error message is bad"""
        message = (
            "The datasource is a required field for SQL file resolvers. "
            "Please add a comment '-- source: my_name'  where 'my_name' refers to a named integration."
        )
        error_builder.add_diagnostic(
            message=message,
            code="143",
            label="missing source",
            range=error_builder.full_comment_range(),
            code_href="https://docs.chalk.ai/docs/integrations",
        )
        errors.append(
            ResolverError(
                display=message,
                path=path,
                parameter=json.dumps(comment_dict),
            )
        )
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=None,
            docstring=docstring,
            errors=errors,
            source=None,
        )
    if "resolves" not in comment_dict:
        """This would be caught by the following BaseModel.parse_obj() but the error message is bad"""
        message = (
            "A feature class must be specified for SQL file resolvers. "
            "Please add a comment '-- resolves: my_name'  where 'my_name' refers to a feature class."
        )
        error_builder.add_diagnostic(
            message=message,
            code="144",
            label="missing feature class: please use the 'resolves' keyword",
            range=error_builder.full_comment_range(),
            code_href="https://docs.chalk.ai/docs/integrations",
        )
        errors.append(
            ResolverError(
                display=message,
                path=path,
                parameter=json.dumps(comment_dict),
            )
        )
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=None,
            docstring=docstring,
            errors=errors,
            source=None,
        )

    try:
        comment_dict_object = CommentDict.parse_obj(comment_dict)
    except ValidationError as e:
        for error in e.errors():
            location = error["loc"][-1]  # the innermost error
            message = f"SQL file resolver could not validate comment '{location}': {error['msg']}"
            range = error_builder.comment_range_by_key(str(location))
            if range is None and location in IncrementalSettingsSQLFileResolver.__fields__:
                range = error_builder.comment_range_by_key("incremental")
            error_builder.add_diagnostic(
                message=message,
                code="145",
                label="Could not validate comment",
                range=range,
            )
            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=json.dumps(comment_dict),
                )
            )
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=None,
            docstring=docstring,
            errors=errors,
            source=None,
        )
    docstring = docstring.strip()

    source_name = comment_dict_object.source
    source = None
    if source_name not in _SOURCES:  # actual name of source
        for possible_source in sources:
            if possible_source.name == source_name:
                source = possible_source
    else:  # source type, e.g. snowflake
        for possible_source in sources:
            source_type = None if source_name is None else _SOURCES.get(source_name)
            if source_type is not None and isinstance(possible_source, source_type):
                if possible_source.name in _SOURCES:
                    source = possible_source
                    break
                if source:
                    message = (
                        f"SQL file resolver refers to '{source_name}' when more than one {source_name}-type source "
                        f"exists. Instead, refer to the integration by name among "
                        f"({[source.name for source in sources]})."
                    )
                    error_builder.add_diagnostic(
                        message=message,
                        code="149",
                        label="Refer to source via name instead",
                        range=error_builder.comment_range_by_key("source"),
                    )
                    errors.append(
                        ResolverError(
                            display=message,
                            path=path,
                            parameter=source_name,
                        )
                    )
                source = possible_source
    if source is None:
        message = (
            f"SQL file resolver refers to unrecognized source '{source_name}'. "
            f"Please refer to your source via its name on your Chalk dashboard, "
            f"and make sure the driver, e.g. chalkpy[snowflake], is installed. "
            f"You will need to instantiate your source somewhere in Python, e.g. "
            f"`source = SnowflakeSource(name='{source_name}')` if it's a Snowflake source."
        )
        error_builder.add_diagnostic(
            message=message,
            code="150",
            label="Source not found",
            range=error_builder.comment_range_by_key("source"),
        )
        errors.append(
            ResolverError(
                display=message,
                path=path,
                parameter=source_name,
            )
        )
    if errors:
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=comment_dict_object,
            docstring=docstring,
            errors=errors,
            source=None,
        )
    assert source is not None, "unrecognized source should be handled by now"
    source_dialect_string = source.get_sqlglot_dialect()
    try:
        glots = sqlglot.parse(sql=sql_string, read=source_dialect_string)
    except Exception as e:
        message = f"SQL file resolver could not SQL parse string: {e}"
        if isinstance(e, sqlglot.errors.ParseError):
            for error in e.errors:
                error_builder.add_diagnostic(
                    message=message,
                    code="146",
                    label="could not parse SQL",
                    range=error_builder.custom_range(
                        line_no=error["line"],
                        col=error["col"] - 2,
                        length=len(error["highlight"]),  # ????
                    ),
                )
        errors.append(ResolverError(display=message, path=path, parameter=None))
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=comment_dict_object,
            docstring=docstring,
            errors=errors,
            source=source,
        )
    if len(glots) > 1:
        message = f"SQL file resolver query {sql_string} has more than one 'SELECT' statements. Only one is permitted."
        error_builder.add_diagnostic(
            message=message,
            code="147",
            label="SQL query must be a single SELECT statement",
            range=error_builder.sql_range(),
        )
        errors.append(ResolverError(display=message, path=path, parameter=None))
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=comment_dict_object,
            docstring=docstring,
            errors=errors,
            source=source,
        )
    glot = glots[0]
    if not isinstance(glot, (sqlglot.expressions.Select, sqlglot.expressions.Union)):
        message = f"SQL file resolver query {sql_string} should be of 'SELECT' type"
        error_builder.add_diagnostic(
            message=message,
            code="147",
            label="SQL query must be a SELECT statement",
            range=error_builder.sql_range(),
        )
        errors.append(ResolverError(display=message, path=path, parameter=None))
        return GlotResult(
            sql_string=sql_string,
            glot=None,
            args=args,
            default_args=default_args,
            comment_dict=comment_dict_object,
            docstring=docstring,
            errors=errors,
            source=source,
        )
    if len(glot.selects) > len(glot.named_selects):
        for select in glot.selects:
            matched = False
            for named_select in glot.named_selects:
                if select.alias_or_name == named_select:
                    matched = True
            if not matched:
                message = (
                    f"SQL file resolver query with unnamed select '{str(select)}'. "
                    f"All selects must either match a feature name, e.g. 'id', or must be aliased, e.g. "
                    f"'SELECT COUNT (DISTINCT merchant_id) AS num_unique_merchant_ids. "
                    f"All names/aliases must match to features on the feature set defined "
                    f"by the 'resolves' comment parameter. "
                )
                error_builder.add_diagnostic(
                    message=message,
                    code="148",
                    label="all selects must map to features",
                    range=error_builder.value_range_by_name(glot=glot, name=str(select).lower()),
                )
                errors.append(
                    ResolverError(
                        display=message,
                        path=path,
                        parameter=None,
                    )
                )
    return GlotResult(
        sql_string=sql_string,
        glot=glot,
        args=args,
        default_args=default_args,
        comment_dict=comment_dict_object,
        docstring=docstring,
        errors=errors,
        source=source,
    )


def _parse_glot(
    glot_result: GlotResult,
    path: str,
    error_builder: SQLFileResolverErrorBuilder,
) -> ParseResult:
    """Parse useful info from sqlglot and gracefully exit if unable to"""
    try:
        import sqlglot
        import sqlglot.expressions
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    # define a source SQL database. Can either specify name or kind if only one of the kind is present.
    comment_dict = glot_result.comment_dict
    assert comment_dict is not None, "comment dict failed to parse"
    docstring = glot_result.docstring
    source = glot_result.source
    errors: List[ResolverError] = []
    # get resolver fields: which columns selected will match to which chalk feature?
    assert comment_dict.resolves is not None, "comment dict failed to parse"
    namespace = build_namespaced_name(namespace=comment_dict.namespace, name=to_snake_case(comment_dict.resolves))
    if namespace not in CURRENT_FEATURE_REGISTRY.get().get_feature_sets():
        message = f"No @features class with the name '{namespace}'"
        error_builder.add_diagnostic(
            message=message,
            code="151",
            label="Unrecognized namespace ",
            range=error_builder.comment_range_by_key("resolves"),
        )
        errors.append(ResolverError(display=message, path=path, parameter=namespace))

    stripped_sql = _remove_comments(glot_result.sql_string)
    lineage = _get_data_lineage(stripped_sql)
    # FIXME: @melrchen: need to make a name for unnamed datasources
    source_name = glot_result.source.name or "" if glot_result.source is not None else ""
    namespaced_lineage = (
        {
            table: {f"{namespace}.{feature}": columns for feature, columns in features.items()}
            for table, features in lineage.items()
        }
        if lineage
        else None
    )
    data_lineage_with_source = {source_name: namespaced_lineage} if namespaced_lineage else None

    if len(errors) > 0:
        return ParseResult(
            sql_string=glot_result.sql_string,
            comment_dict=comment_dict,
            fields={},
            namespace=namespace,
            source=source,
            docstring=docstring,
            data_lineage=data_lineage_with_source,
            errors=errors,
        )

    assert isinstance(
        glot_result.glot, (sqlglot.expressions.Select, sqlglot.expressions.Union)
    ), f"glot {glot_result.glot} is not a select or union statement"
    # sql target -> output feature string
    fields: Dict[str, str] = {
        column_name: f"{namespace}.{column_name}" for column_name in glot_result.glot.named_selects
    }

    return ParseResult(
        sql_string=glot_result.sql_string,
        comment_dict=comment_dict,
        fields=fields,
        namespace=namespace,
        source=source,
        docstring=docstring,
        data_lineage=data_lineage_with_source,
        errors=errors,
    )


def _get_stream_resolver(
    path: str,
    glot_result: GlotResult,
    parsed: ParseResult,
    outputs: List[Feature],
    error_builder: SQLFileResolverErrorBuilder,
) -> ResolverResult:
    errors = []
    output_features = Features[DataFrame[tuple(outputs)]]

    if isinstance(output_features.features[0], type) and issubclass(output_features.features[0], DataFrame):
        output_feature_fqns = set(f.fqn for f in cast(Type[DataFrame], output_features.features[0]).columns)
    else:
        output_feature_fqns = set(f.fqn for f in output_features.features)

    signature = StreamResolverSignature(
        params=[],
        output_feature_fqns=output_feature_fqns,
    )

    sql_query: str = _remove_comments(parsed.sql_string)
    filename = os.path.basename(path)
    try:

        def fn():
            return sql_query

        # attempt to instantiate the resolver
        resolver = StreamResolver(
            function_definition=sql_query,
            fqn=filename.replace(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX, ""),
            filename=path,
            source=cast(StreamSource, parsed.source),
            fn=fn,
            environment=parsed.comment_dict.environment,
            doc=parsed.docstring,
            mode=None,
            machine_type=parsed.comment_dict.machine_type,
            message=None,
            output=output_features,
            signature=signature,
            state=None,
            sql_query=sql_query,
            owner=parsed.comment_dict.owner,
            parse=None,
            keys=None,  # TODO implement parse and keys for sql file resolvers?
            timestamp=None,
            source_line=None,
            tags=None,
            lsp_builder=get_resolver_error_builder(fn),
            autogenerated=False,
            updates_materialized_aggregations=True,
            sql_settings=None,
        )
    except Exception as e:
        message = f"Streaming SQL file resolver could not be instantiated, {e}"
        error_builder.add_diagnostic(
            message=message,
            code="141",
            label="resolver instantiation failed",
            range=error_builder.full_range(),
        )
        errors.append(
            ResolverError(
                display=message,
                path=path,
                parameter=None,
            )
        )
        return ResolverResult(resolver=None, errors=errors, db=parsed.source, fields=None, args=None)

    return ResolverResult(
        resolver=resolver,
        errors=errors,
        db=parsed.source,
        fields=parsed.fields,
        args=glot_result.args,
    )


def _validate_feature_strings_in_comments(
    feature_strings: list[str],
    error_builder: SQLFileResolverErrorBuilder,
    namespace: str,
    path: str,
    errors: list[ResolverError],
    outputs: list[Feature],
    arg_name: str,
) -> tuple[Feature, ...]:
    feature_list: list[Feature] = []
    for f in feature_strings:
        feature = None
        try:
            feature = Feature.from_root_fqn(f)
        except:
            if "." not in f:
                reconstructed_fqn = f"{namespace}.{f}"
                try:
                    feature = Feature.from_root_fqn(reconstructed_fqn)
                except:
                    pass
        if feature is None:
            message = f"SQL file resolver {path} references a 'unique_on' feature '{f}' which does not exist."

            error_builder.add_diagnostic_with_spellcheck(
                spellcheck_item=f,
                spellcheck_candidates=[
                    feature.fqn
                    for feature in CURRENT_FEATURE_REGISTRY.get()
                    .get_feature_sets()[build_namespaced_name(name=namespace)]
                    .features
                ],
                message=message,
                code="153a",
                label="output feature not recognized",
                range=error_builder.comment_range_by_key(arg_name),
            )
            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=f,
                )
            )
            continue

        if feature not in outputs:
            message = (
                f"SQL file resolver {path} references a '{arg_name}' feature '{f}' which is not an output feature."
            )
            error_builder.add_diagnostic(
                message=message,
                code="153b",
                label=f"{arg_name} feature not in outputs",
                range=error_builder.comment_range_by_key(arg_name),
            )
            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=f,
                )
            )
        elif feature in feature_list:
            message = f"SQL file resolver {path} references '{arg_name}' feature '{f}' multiple times."
            error_builder.add_diagnostic(
                message=message,
                code="153b",
                label=f"{arg_name} feature stated multiple times",
                range=error_builder.comment_range_by_key(arg_name),
            )
            errors.append(
                ResolverError(
                    display=message,
                    path=path,
                    parameter=f,
                )
            )
        else:
            feature_list.append(feature)
    return tuple(feature_list)


def _remove_comments(sql_string: str) -> str:
    sql_string = re.sub(
        re.compile("/\\*.*?\\*/", re.DOTALL), "", sql_string
    )  # remove all occurrences streamed comments (/*COMMENT */) from string
    sql_string = re.sub(
        re.compile("//.*?\n"), "", sql_string
    )  # remove all occurrence single-line comments (//COMMENT\n ) from string
    sql_string = re.sub(
        re.compile("--.*?\n"), "", sql_string
    )  # remove all occurrence single-line comments (//COMMENT\n ) from string
    return sql_string.strip()


def count_colons(s: str) -> int:
    count = 0
    inside_single_quotes = False
    inside_double_quotes = False

    for char in s:
        if char == '"':
            inside_double_quotes = not inside_double_quotes
        elif char == "'":
            inside_single_quotes = not inside_single_quotes
        elif char == ":" and not inside_single_quotes and not inside_double_quotes:
            count += 1
    return count


@dataclasses.dataclass(frozen=True)
class GeneratedSQLFileResolverInfo:
    filepath: str
    sql_string: str
    comment_dict: CommentDict


class GeneratedSQLFileResolverRegistry:
    def __init__(self):
        super().__init__()
        self.resolver_name_to_generated_infos: Dict[str, GeneratedSQLFileResolverInfo] = {}

    def add_sql_file_resolver(self, name: str, filepath: str, sql_string: str, comment_dict: CommentDict):
        if name in self.resolver_name_to_generated_infos and filepath != "<notebook>":
            raise ValueError(f"A SQL file resolver already exists with name '{name}'. They must have unique names.")
        self.resolver_name_to_generated_infos[name] = GeneratedSQLFileResolverInfo(
            filepath=filepath,
            sql_string=sql_string,
            comment_dict=comment_dict,
        )

    def get_generated_sql_file_resolvers(self) -> Iterable[SQLStringResult]:
        for name, generated_info in self.resolver_name_to_generated_infos.items():
            yield SQLStringResult(
                path=generated_info.filepath,
                sql_string=generated_info.sql_string,
                error=None,
                override_comment_dict=generated_info.comment_dict,
                override_name=name,
                autogenerated=True,
            )


_GENERATED_SQL_FILE_RESOLVER_REGISTRY = GeneratedSQLFileResolverRegistry()

# name --> resolver
NOTEBOOK_DEFINED_SQL_RESOLVERS: Dict[str, ResolverResult] = {}


def make_sql_file_resolver(
    name: str,
    sql: str,
    source: str | BaseSQLSourceProtocol | None = None,
    resolves: str | Any | None = None,
    kind: Literal["online", "offline", "streaming"] | None = None,
    incremental: IncrementalSettings | None = None,
    count: Literal[1, "one", "one_or_none", "all"] | None = None,
    timeout: Duration | None = None,
    cron: CronTab | Duration | Cron | None = None,
    owner: str | None = None,
    machine_type: MachineType | None = None,
    fields: dict[str, str | FeatureReference] | None = None,
    environment: Environments | None = None,
    tags: Tags | None = None,
    unique_on: Collection[FeatureReference] | None = None,
    partitioned_by: Collection[Any] | None = None,
    total: Optional[bool] = None,
):
    """Generate a Chalk SQL file resolver from a filepath and a sql string.
    This will generate a resolver in your web dashboard that can be queried,
    but will not output a `.chalk.sql` file.

    The optional parameters are overrides for the comment key-value pairs
    at the top of the sql file resolver. Comment key-value pairs specify
    important resolver information such as the source, feature namespace
    to resolve, and other details.
    Note that these will override any values specified in the sql string.
    See https://docs.chalk.ai/docs/sql#configuration for more information.

    See https://docs.chalk.ai/docs/sql#sql-file-resolvers for more information
    on SQL file resolvers.

    Parameters
    ----------
    name
        The name of your resolver
    sql
        The sql string for your query.
    kind
        The type of resolver.
        If not specified, defaults to "online".
    source
        Can either be a BaseSQLSource or a string.
        If a string is provided, it will be used to infer the source by
        first scanning for a source with the same name, then inferring
        the source if it is a type, e.g. `snowflake` if there is only
        one database of that type. Optional if `source` is specified in `sql`.
    resolves
        Describes the feature namespace to which the outputs belong.
        Optional if `resolves` is specified in `sql`.

    Other Parameters
    ----------------
    incremental
        Parameters for incremental queries.
        For more information, see https://docs.chalk.ai/docs/sql#incremental-queries.
    count
        If set to `one`, the resolver will return only one row.
        If set to `one_or_none`, the resolver will return at most one row.
        If set to `all`, the resolver will return all rows
    timeout
        You can specify the maximum duration to wait for the
        resolver's result. Once the resolver's runtime exceeds
        the specified duration, a timeout error will be returned
        along with each output feature.

        Please use supported Chalk durations
        'w', 'd', 'h', 'm', 's', and/or 'ms'.

        Read more at https://docs.chalk.ai/docs/timeout and https://docs.chalk.ai/api-docs#Duration
    cron
        You can schedule resolvers to run on a pre-determined
        schedule via the cron argument to resolver decorators.

        Cron can sample all examples, a subset of all examples,
        or a custom provided set of examples.

        Read more at https://docs.chalk.ai/docs/resolver-cron
    environment
        Environments are used to trigger behavior
        in different deployments such as staging, production, and
        local development. For example, you may wish to interact with
        a vendor via an API call in the production environment, and
        opt to return a constant value in a staging environment.

        Environment can take one of three types:
            - `None` (default) - candidate to run in every environment
            - `str` - run only in this environment
            - `list[str]` - run in any of the specified environment and no others

        Read more at https://docs.chalk.ai/docs/resolver-environments
    tags
        Allow you to scope requests within an
        environment. Both tags and environment need to match for a
        resolver to be a candidate to execute.

        You might consider using tags, for example, to change out
        whether you want to use a sandbox environment for a vendor,
        or to bypass the vendor and return constant values in a
        staging environment.

        Read more at https://docs.chalk.ai/docs/resolver-tags
    owner
        Individual or team responsible for this resolver.
        The Chalk Dashboard will display this field, and alerts
        can be routed to owners.
    machine_type
        You can optionally specify that resolvers need to run on
        a machine other than the default. Must be configured in
        your deployment.
    fields
        An optional mapping from SQL column to Chalk feature.
        For example, let's say we have a Chalk feature class 'Transaction' with a primary key 'id'.
        If we have a SQL query like `SELECT txn_id from transactions`,
        we can map the `txn_id` to our Chalk feature with the mapping
        `txn_id: id`.
    unique_on
        A list of features that must be unique for each row of the output.
        This enables unique optimizations in the resolver execution.
        Only applicable to resolvers that return a DataFrame.
    partitioned_by
        A list of features that correspond to partition keys in the data source.
        This field indicates that this resolver executes its query against a data storage system that is
        partitioned by a particular set of columns.
        This is most common with data-warehouse sources like Snowflake, BigQuery or Databricks.
    total
        Whether this resolver returns all ids of a given namespace. To have this annotation, the resolver must
        take no arguments and return a DataFrame.


    Examples
    --------
    >>> from chalk import make_sql_file_resolver
    >>> from chalk.features import features
    >>> @features
    ... class User:
    ...     id: int
    ...     name: str
    >>> make_sql_file_resolver(
    ...     name="my_resolver",
    ...     sql="SELECT user_id as id, name FROM users",
    ...     source="snowflake",
    ...     resolves=User,
    ...     kind="offline",
    ... )
    """
    if name.endswith(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX):
        name = name.replace(CHALK_SQL_FILE_RESOLVER_FILENAME_SUFFIX, "")
    filename = None
    frame = inspect.currentframe()
    assert frame is not None, "Failed to get current frame"
    caller_frame = frame.f_back
    assert caller_frame is not None, "Failed to get caller frame"
    filename = caller_frame.f_code.co_filename
    del frame
    is_defined_in_notebook: bool = False
    if notebook.is_notebook():
        module = inspect.getmodule(caller_frame)
        if module is not None:
            filename = module.__name__
        else:
            filename = "<notebook>"
            is_defined_in_notebook = True

    comment_dict = CommentDict(
        source=source if isinstance(source, str) or source is None else source.name,
        resolves=str(resolves) if resolves is not None else None,
        type=kind,
        incremental=incremental and _convert_incremental_settings(incremental),
        count=count,
        timeout=(
            timeout
            if timeout is None
            else timedelta_to_duration(timeout)
            if isinstance(timeout, timedelta)
            else timeout
        ),
        cron=cron,
        owner=owner,
        fields=None if fields is None else {k: str(v) for k, v in fields.items()},
        environment=(
            None if environment is None else [environment] if isinstance(environment, str) else list(environment)
        ),
        tags=None if tags is None else [tags] if isinstance(tags, str) else list(tags),
        total=total,
        namespace=None,
        machine_type=machine_type,
        unique_on=[str(f) for f in unique_on] if unique_on is not None else None,
        partitioned_by=[str(f) for f in partitioned_by] if partitioned_by is not None else None,
    )
    _GENERATED_SQL_FILE_RESOLVER_REGISTRY.add_sql_file_resolver(
        filepath=filename,
        sql_string=sql,
        comment_dict=comment_dict,
        name=name,
    )
    if is_defined_in_notebook:
        from chalk.sql import SQLSourceGroup

        current_sql_sources: List[BaseSQLSource | SQLSourceGroup] = [
            *BaseSQLSource.registry,
            *SQLSourceGroup.registry,
        ]
        if isinstance(source, str):
            source_names = {s.name for s in current_sql_sources}
            if source not in _SOURCES:
                if source not in source_names:
                    msg = f"Unable to create SQL resolver '{name}' since a SQL source with the name '{source}' was not found. Please make sure the SQL source exists & is imported in the current environment."
                    if len(source_names) <= 10:
                        msg += f" Currently loaded SQL sources: {source_names}"
                    raise RuntimeError(msg)
            else:
                source_type = _SOURCES[source]
                possible_sources = [s for s in current_sql_sources if isinstance(s, source_type)]
                if len(possible_sources) == 0:
                    msg = f"Unable to create SQL resolver '{name}' since no SQL sources of type '{source}' were found. Please make sure the SQL source exists & is imported in the current environment."
                    if len(source_names) <= 10:
                        msg += f" Currently loaded SQL sources: {source_names}"
                    raise RuntimeError(msg)
                elif len(possible_sources) > 1:
                    msg = f"Unable to create SQL resolver '{name}' since multiple SQL sources of type '{source}' were found: {[x.name for x in possible_sources]} Please refer to the SQL source by name (e.g. '{possible_sources[0].name}') instead of by type."
                    raise RuntimeError(msg)

        generated_info = _GENERATED_SQL_FILE_RESOLVER_REGISTRY.resolver_name_to_generated_infos[name]
        info = SQLStringResult(
            path=generated_info.filepath,
            sql_string=generated_info.sql_string,
            error=None,
            override_comment_dict=generated_info.comment_dict,
            override_name=name,
            autogenerated=True,
        )
        resolver_result = get_sql_file_resolver(
            sources=current_sql_sources, sql_string_result=info, has_import_errors=False
        )
        if resolver_result.errors:
            errs = [e.display for e in resolver_result.errors]
            err_message = "\n".join(errs)
            raise RuntimeError(
                f"Failed to parse notebook-defined SQL resolver '{name}'. Found the following errors:\n{err_message}"
            )
        NOTEBOOK_DEFINED_SQL_RESOLVERS[name] = resolver_result
        return resolver_result


def _convert_incremental_settings(settings: IncrementalSettings) -> IncrementalSettingsSQLFileResolver:
    return IncrementalSettingsSQLFileResolver(
        incremental_column=settings.incremental_column,
        lookback_period=(
            timedelta_to_duration(settings.lookback_period) if settings.lookback_period is not None else None
        ),
        mode=settings.mode,
        incremental_timestamp=settings.incremental_timestamp,
    )


def parse_finalizer(count: Literal[1, "one", "one_or_none", "all"] | None) -> Finalizer | None:
    if count == 1 or count == "one":
        return Finalizer.ONE
    if count == "one_or_none":
        return Finalizer.ONE_OR_NONE
    if count == "all":
        return Finalizer.ALL
    return None



from google.protobuf import duration_pb2 as _duration_pb2
from google.rpc import code_pb2 as _code_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class NodePoolUpdateStrategy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    NODE_POOL_UPDATE_STRATEGY_UNSPECIFIED: _ClassVar[NodePoolUpdateStrategy]
    NODE_POOL_UPDATE_STRATEGY_BLUE_GREEN: _ClassVar[NodePoolUpdateStrategy]
    NODE_POOL_UPDATE_STRATEGY_SURGE: _ClassVar[NodePoolUpdateStrategy]

NODE_POOL_UPDATE_STRATEGY_UNSPECIFIED: NodePoolUpdateStrategy
NODE_POOL_UPDATE_STRATEGY_BLUE_GREEN: NodePoolUpdateStrategy
NODE_POOL_UPDATE_STRATEGY_SURGE: NodePoolUpdateStrategy

class StatusCondition(_message.Message):
    __slots__ = ("code", "message", "canonical_code")
    class Code(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        CODE_UNSPECIFIED: _ClassVar[StatusCondition.Code]
        CODE_GCE_STOCKOUT: _ClassVar[StatusCondition.Code]
        CODE_GKE_SERVICE_ACCOUNT_DELETED: _ClassVar[StatusCondition.Code]
        CODE_GCE_QUOTA_EXCEEDED: _ClassVar[StatusCondition.Code]
        CODE_SET_BY_OPERATOR: _ClassVar[StatusCondition.Code]
        CODE_CLOUD_KMS_KEY_ERROR: _ClassVar[StatusCondition.Code]
        CODE_CA_EXPIRING: _ClassVar[StatusCondition.Code]
        CODE_NODE_SERVICE_ACCOUNT_MISSING_PERMISSIONS: _ClassVar[StatusCondition.Code]

    CODE_UNSPECIFIED: StatusCondition.Code
    CODE_GCE_STOCKOUT: StatusCondition.Code
    CODE_GKE_SERVICE_ACCOUNT_DELETED: StatusCondition.Code
    CODE_GCE_QUOTA_EXCEEDED: StatusCondition.Code
    CODE_SET_BY_OPERATOR: StatusCondition.Code
    CODE_CLOUD_KMS_KEY_ERROR: StatusCondition.Code
    CODE_CA_EXPIRING: StatusCondition.Code
    CODE_NODE_SERVICE_ACCOUNT_MISSING_PERMISSIONS: StatusCondition.Code
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    CANONICAL_CODE_FIELD_NUMBER: _ClassVar[int]
    code: StatusCondition.Code
    message: str
    canonical_code: _code_pb2.Code
    def __init__(
        self,
        code: _Optional[_Union[StatusCondition.Code, str]] = ...,
        message: _Optional[str] = ...,
        canonical_code: _Optional[_Union[_code_pb2.Code, str]] = ...,
    ) -> None: ...

class NodePoolAutoscaling(_message.Message):
    __slots__ = (
        "enabled",
        "min_node_count",
        "max_node_count",
        "autoprovisioned",
        "location_policy",
        "total_min_node_count",
        "total_max_node_count",
    )
    class LocationPolicy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        LOCATION_POLICY_UNSPECIFIED: _ClassVar[NodePoolAutoscaling.LocationPolicy]
        LOCATION_POLICY_BALANCED: _ClassVar[NodePoolAutoscaling.LocationPolicy]
        LOCATION_POLICY_ANY: _ClassVar[NodePoolAutoscaling.LocationPolicy]

    LOCATION_POLICY_UNSPECIFIED: NodePoolAutoscaling.LocationPolicy
    LOCATION_POLICY_BALANCED: NodePoolAutoscaling.LocationPolicy
    LOCATION_POLICY_ANY: NodePoolAutoscaling.LocationPolicy
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    MIN_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
    MAX_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
    AUTOPROVISIONED_FIELD_NUMBER: _ClassVar[int]
    LOCATION_POLICY_FIELD_NUMBER: _ClassVar[int]
    TOTAL_MIN_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
    TOTAL_MAX_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
    enabled: bool
    min_node_count: int
    max_node_count: int
    autoprovisioned: bool
    location_policy: NodePoolAutoscaling.LocationPolicy
    total_min_node_count: int
    total_max_node_count: int
    def __init__(
        self,
        enabled: bool = ...,
        min_node_count: _Optional[int] = ...,
        max_node_count: _Optional[int] = ...,
        autoprovisioned: bool = ...,
        location_policy: _Optional[_Union[NodePoolAutoscaling.LocationPolicy, str]] = ...,
        total_min_node_count: _Optional[int] = ...,
        total_max_node_count: _Optional[int] = ...,
    ) -> None: ...

class NodeTaint(_message.Message):
    __slots__ = ("key", "value", "effect")
    class Effect(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        EFFECT_UNSPECIFIED: _ClassVar[NodeTaint.Effect]
        EFFECT_NO_SCHEDULE: _ClassVar[NodeTaint.Effect]
        EFFECT_PREFER_NO_SCHEDULE: _ClassVar[NodeTaint.Effect]
        EFFECT_NO_EXECUTE: _ClassVar[NodeTaint.Effect]

    EFFECT_UNSPECIFIED: NodeTaint.Effect
    EFFECT_NO_SCHEDULE: NodeTaint.Effect
    EFFECT_PREFER_NO_SCHEDULE: NodeTaint.Effect
    EFFECT_NO_EXECUTE: NodeTaint.Effect
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    EFFECT_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    effect: NodeTaint.Effect
    def __init__(
        self,
        key: _Optional[str] = ...,
        value: _Optional[str] = ...,
        effect: _Optional[_Union[NodeTaint.Effect, str]] = ...,
    ) -> None: ...

class MaxPodsConstraint(_message.Message):
    __slots__ = ("max_pods_per_node",)
    MAX_PODS_PER_NODE_FIELD_NUMBER: _ClassVar[int]
    max_pods_per_node: int
    def __init__(self, max_pods_per_node: _Optional[int] = ...) -> None: ...

class BestEffortProvisioning(_message.Message):
    __slots__ = ("enabled", "min_provision_nodes")
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    MIN_PROVISION_NODES_FIELD_NUMBER: _ClassVar[int]
    enabled: bool
    min_provision_nodes: int
    def __init__(self, enabled: bool = ..., min_provision_nodes: _Optional[int] = ...) -> None: ...

class NodeManagement(_message.Message):
    __slots__ = ("auto_upgrade", "auto_repair", "upgrade_options")
    AUTO_UPGRADE_FIELD_NUMBER: _ClassVar[int]
    AUTO_REPAIR_FIELD_NUMBER: _ClassVar[int]
    UPGRADE_OPTIONS_FIELD_NUMBER: _ClassVar[int]
    auto_upgrade: bool
    auto_repair: bool
    upgrade_options: AutoUpgradeOptions
    def __init__(
        self,
        auto_upgrade: bool = ...,
        auto_repair: bool = ...,
        upgrade_options: _Optional[_Union[AutoUpgradeOptions, _Mapping]] = ...,
    ) -> None: ...

class AutoUpgradeOptions(_message.Message):
    __slots__ = ("auto_upgrade_start_time", "description")
    AUTO_UPGRADE_START_TIME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    auto_upgrade_start_time: str
    description: str
    def __init__(self, auto_upgrade_start_time: _Optional[str] = ..., description: _Optional[str] = ...) -> None: ...

class NodeConfig(_message.Message):
    __slots__ = (
        "machine_type",
        "disk_size_gb",
        "oauth_scopes",
        "service_account",
        "metadata",
        "image_type",
        "labels",
        "local_ssd_count",
        "tags",
        "preemptible",
        "disk_type",
        "min_cpu_platform",
        "taints",
        "node_group",
        "boot_disk_kms_key",
        "spot",
        "resource_labels",
        "local_ssd_encryption_mode",
        "effective_cgroup_mode",
    )
    class LocalSsdEncryptionMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        LOCAL_SSD_ENCRYPTION_MODE_UNSPECIFIED: _ClassVar[NodeConfig.LocalSsdEncryptionMode]
        LOCAL_SSD_ENCRYPTION_MODE_STANDARD_ENCRYPTION: _ClassVar[NodeConfig.LocalSsdEncryptionMode]
        LOCAL_SSD_ENCRYPTION_MODE_EPHEMERAL_KEY_ENCRYPTION: _ClassVar[NodeConfig.LocalSsdEncryptionMode]

    LOCAL_SSD_ENCRYPTION_MODE_UNSPECIFIED: NodeConfig.LocalSsdEncryptionMode
    LOCAL_SSD_ENCRYPTION_MODE_STANDARD_ENCRYPTION: NodeConfig.LocalSsdEncryptionMode
    LOCAL_SSD_ENCRYPTION_MODE_EPHEMERAL_KEY_ENCRYPTION: NodeConfig.LocalSsdEncryptionMode
    class EffectiveCgroupMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        EFFECTIVE_CGROUP_MODE_UNSPECIFIED: _ClassVar[NodeConfig.EffectiveCgroupMode]
        EFFECTIVE_CGROUP_MODE_V1: _ClassVar[NodeConfig.EffectiveCgroupMode]
        EFFECTIVE_CGROUP_MODE_V2: _ClassVar[NodeConfig.EffectiveCgroupMode]

    EFFECTIVE_CGROUP_MODE_UNSPECIFIED: NodeConfig.EffectiveCgroupMode
    EFFECTIVE_CGROUP_MODE_V1: NodeConfig.EffectiveCgroupMode
    EFFECTIVE_CGROUP_MODE_V2: NodeConfig.EffectiveCgroupMode
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class LabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    class ResourceLabelsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

    MACHINE_TYPE_FIELD_NUMBER: _ClassVar[int]
    DISK_SIZE_GB_FIELD_NUMBER: _ClassVar[int]
    OAUTH_SCOPES_FIELD_NUMBER: _ClassVar[int]
    SERVICE_ACCOUNT_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    IMAGE_TYPE_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    LOCAL_SSD_COUNT_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    PREEMPTIBLE_FIELD_NUMBER: _ClassVar[int]
    DISK_TYPE_FIELD_NUMBER: _ClassVar[int]
    MIN_CPU_PLATFORM_FIELD_NUMBER: _ClassVar[int]
    TAINTS_FIELD_NUMBER: _ClassVar[int]
    NODE_GROUP_FIELD_NUMBER: _ClassVar[int]
    BOOT_DISK_KMS_KEY_FIELD_NUMBER: _ClassVar[int]
    SPOT_FIELD_NUMBER: _ClassVar[int]
    RESOURCE_LABELS_FIELD_NUMBER: _ClassVar[int]
    LOCAL_SSD_ENCRYPTION_MODE_FIELD_NUMBER: _ClassVar[int]
    EFFECTIVE_CGROUP_MODE_FIELD_NUMBER: _ClassVar[int]
    machine_type: str
    disk_size_gb: int
    oauth_scopes: _containers.RepeatedScalarFieldContainer[str]
    service_account: str
    metadata: _containers.ScalarMap[str, str]
    image_type: str
    labels: _containers.ScalarMap[str, str]
    local_ssd_count: int
    tags: _containers.RepeatedScalarFieldContainer[str]
    preemptible: bool
    disk_type: str
    min_cpu_platform: str
    taints: _containers.RepeatedCompositeFieldContainer[NodeTaint]
    node_group: str
    boot_disk_kms_key: str
    spot: bool
    resource_labels: _containers.ScalarMap[str, str]
    local_ssd_encryption_mode: NodeConfig.LocalSsdEncryptionMode
    effective_cgroup_mode: NodeConfig.EffectiveCgroupMode
    def __init__(
        self,
        machine_type: _Optional[str] = ...,
        disk_size_gb: _Optional[int] = ...,
        oauth_scopes: _Optional[_Iterable[str]] = ...,
        service_account: _Optional[str] = ...,
        metadata: _Optional[_Mapping[str, str]] = ...,
        image_type: _Optional[str] = ...,
        labels: _Optional[_Mapping[str, str]] = ...,
        local_ssd_count: _Optional[int] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        preemptible: bool = ...,
        disk_type: _Optional[str] = ...,
        min_cpu_platform: _Optional[str] = ...,
        taints: _Optional[_Iterable[_Union[NodeTaint, _Mapping]]] = ...,
        node_group: _Optional[str] = ...,
        boot_disk_kms_key: _Optional[str] = ...,
        spot: bool = ...,
        resource_labels: _Optional[_Mapping[str, str]] = ...,
        local_ssd_encryption_mode: _Optional[_Union[NodeConfig.LocalSsdEncryptionMode, str]] = ...,
        effective_cgroup_mode: _Optional[_Union[NodeConfig.EffectiveCgroupMode, str]] = ...,
    ) -> None: ...

class BlueGreenSettings(_message.Message):
    __slots__ = ("standard_rollout_policy", "node_pool_soak_duration")
    class StandardRolloutPolicy(_message.Message):
        __slots__ = ("batch_percentage", "batch_node_count", "batch_soak_duration")
        BATCH_PERCENTAGE_FIELD_NUMBER: _ClassVar[int]
        BATCH_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
        BATCH_SOAK_DURATION_FIELD_NUMBER: _ClassVar[int]
        batch_percentage: float
        batch_node_count: int
        batch_soak_duration: _duration_pb2.Duration
        def __init__(
            self,
            batch_percentage: _Optional[float] = ...,
            batch_node_count: _Optional[int] = ...,
            batch_soak_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
        ) -> None: ...

    STANDARD_ROLLOUT_POLICY_FIELD_NUMBER: _ClassVar[int]
    NODE_POOL_SOAK_DURATION_FIELD_NUMBER: _ClassVar[int]
    standard_rollout_policy: BlueGreenSettings.StandardRolloutPolicy
    node_pool_soak_duration: _duration_pb2.Duration
    def __init__(
        self,
        standard_rollout_policy: _Optional[_Union[BlueGreenSettings.StandardRolloutPolicy, _Mapping]] = ...,
        node_pool_soak_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ...,
    ) -> None: ...

class GKENodePool(_message.Message):
    __slots__ = (
        "name",
        "config",
        "initial_node_count",
        "locations",
        "self_link",
        "version",
        "instance_group_urls",
        "status",
        "status_message",
        "autoscaling",
        "management",
        "max_pods_constraint",
        "conditions",
        "pod_ipv4_cidr_size",
        "upgrade_settings",
        "placement_policy",
        "update_info",
        "etag",
        "queued_provisioning",
        "best_effort_provisioning",
    )
    class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATUS_UNSPECIFIED: _ClassVar[GKENodePool.Status]
        STATUS_PROVISIONING: _ClassVar[GKENodePool.Status]
        STATUS_RUNNING: _ClassVar[GKENodePool.Status]
        STATUS_RUNNING_WITH_ERROR: _ClassVar[GKENodePool.Status]
        STATUS_RECONCILING: _ClassVar[GKENodePool.Status]
        STATUS_STOPPING: _ClassVar[GKENodePool.Status]
        STATUS_ERROR: _ClassVar[GKENodePool.Status]

    STATUS_UNSPECIFIED: GKENodePool.Status
    STATUS_PROVISIONING: GKENodePool.Status
    STATUS_RUNNING: GKENodePool.Status
    STATUS_RUNNING_WITH_ERROR: GKENodePool.Status
    STATUS_RECONCILING: GKENodePool.Status
    STATUS_STOPPING: GKENodePool.Status
    STATUS_ERROR: GKENodePool.Status
    class UpgradeSettings(_message.Message):
        __slots__ = ("max_surge", "max_unavailable", "strategy", "blue_green_settings")
        MAX_SURGE_FIELD_NUMBER: _ClassVar[int]
        MAX_UNAVAILABLE_FIELD_NUMBER: _ClassVar[int]
        STRATEGY_FIELD_NUMBER: _ClassVar[int]
        BLUE_GREEN_SETTINGS_FIELD_NUMBER: _ClassVar[int]
        max_surge: int
        max_unavailable: int
        strategy: NodePoolUpdateStrategy
        blue_green_settings: BlueGreenSettings
        def __init__(
            self,
            max_surge: _Optional[int] = ...,
            max_unavailable: _Optional[int] = ...,
            strategy: _Optional[_Union[NodePoolUpdateStrategy, str]] = ...,
            blue_green_settings: _Optional[_Union[BlueGreenSettings, _Mapping]] = ...,
        ) -> None: ...

    class UpdateInfo(_message.Message):
        __slots__ = ("blue_green_info",)
        class BlueGreenInfo(_message.Message):
            __slots__ = (
                "phase",
                "blue_instance_group_urls",
                "green_instance_group_urls",
                "blue_pool_deletion_start_time",
                "green_pool_version",
            )
            class Phase(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
                __slots__ = ()
                PHASE_UNSPECIFIED: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_UPDATE_STARTED: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_CREATING_GREEN_POOL: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_CORDONING_BLUE_POOL: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_DRAINING_BLUE_POOL: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_NODE_POOL_SOAKING: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_DELETING_BLUE_POOL: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]
                PHASE_ROLLBACK_STARTED: _ClassVar[GKENodePool.UpdateInfo.BlueGreenInfo.Phase]

            PHASE_UNSPECIFIED: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_UPDATE_STARTED: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_CREATING_GREEN_POOL: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_CORDONING_BLUE_POOL: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_DRAINING_BLUE_POOL: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_NODE_POOL_SOAKING: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_DELETING_BLUE_POOL: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_ROLLBACK_STARTED: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            PHASE_FIELD_NUMBER: _ClassVar[int]
            BLUE_INSTANCE_GROUP_URLS_FIELD_NUMBER: _ClassVar[int]
            GREEN_INSTANCE_GROUP_URLS_FIELD_NUMBER: _ClassVar[int]
            BLUE_POOL_DELETION_START_TIME_FIELD_NUMBER: _ClassVar[int]
            GREEN_POOL_VERSION_FIELD_NUMBER: _ClassVar[int]
            phase: GKENodePool.UpdateInfo.BlueGreenInfo.Phase
            blue_instance_group_urls: _containers.RepeatedScalarFieldContainer[str]
            green_instance_group_urls: _containers.RepeatedScalarFieldContainer[str]
            blue_pool_deletion_start_time: str
            green_pool_version: str
            def __init__(
                self,
                phase: _Optional[_Union[GKENodePool.UpdateInfo.BlueGreenInfo.Phase, str]] = ...,
                blue_instance_group_urls: _Optional[_Iterable[str]] = ...,
                green_instance_group_urls: _Optional[_Iterable[str]] = ...,
                blue_pool_deletion_start_time: _Optional[str] = ...,
                green_pool_version: _Optional[str] = ...,
            ) -> None: ...

        BLUE_GREEN_INFO_FIELD_NUMBER: _ClassVar[int]
        blue_green_info: GKENodePool.UpdateInfo.BlueGreenInfo
        def __init__(
            self, blue_green_info: _Optional[_Union[GKENodePool.UpdateInfo.BlueGreenInfo, _Mapping]] = ...
        ) -> None: ...

    class PlacementPolicy(_message.Message):
        __slots__ = ("type", "tpu_topology", "policy_name")
        class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
            __slots__ = ()
            TYPE_UNSPECIFIED: _ClassVar[GKENodePool.PlacementPolicy.Type]
            TYPE_COMPACT: _ClassVar[GKENodePool.PlacementPolicy.Type]

        TYPE_UNSPECIFIED: GKENodePool.PlacementPolicy.Type
        TYPE_COMPACT: GKENodePool.PlacementPolicy.Type
        TYPE_FIELD_NUMBER: _ClassVar[int]
        TPU_TOPOLOGY_FIELD_NUMBER: _ClassVar[int]
        POLICY_NAME_FIELD_NUMBER: _ClassVar[int]
        type: GKENodePool.PlacementPolicy.Type
        tpu_topology: str
        policy_name: str
        def __init__(
            self,
            type: _Optional[_Union[GKENodePool.PlacementPolicy.Type, str]] = ...,
            tpu_topology: _Optional[str] = ...,
            policy_name: _Optional[str] = ...,
        ) -> None: ...

    class QueuedProvisioning(_message.Message):
        __slots__ = ("enabled",)
        ENABLED_FIELD_NUMBER: _ClassVar[int]
        enabled: bool
        def __init__(self, enabled: bool = ...) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    INITIAL_NODE_COUNT_FIELD_NUMBER: _ClassVar[int]
    LOCATIONS_FIELD_NUMBER: _ClassVar[int]
    SELF_LINK_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_GROUP_URLS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STATUS_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    AUTOSCALING_FIELD_NUMBER: _ClassVar[int]
    MANAGEMENT_FIELD_NUMBER: _ClassVar[int]
    MAX_PODS_CONSTRAINT_FIELD_NUMBER: _ClassVar[int]
    CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    POD_IPV4_CIDR_SIZE_FIELD_NUMBER: _ClassVar[int]
    UPGRADE_SETTINGS_FIELD_NUMBER: _ClassVar[int]
    PLACEMENT_POLICY_FIELD_NUMBER: _ClassVar[int]
    UPDATE_INFO_FIELD_NUMBER: _ClassVar[int]
    ETAG_FIELD_NUMBER: _ClassVar[int]
    QUEUED_PROVISIONING_FIELD_NUMBER: _ClassVar[int]
    BEST_EFFORT_PROVISIONING_FIELD_NUMBER: _ClassVar[int]
    name: str
    config: NodeConfig
    initial_node_count: int
    locations: _containers.RepeatedScalarFieldContainer[str]
    self_link: str
    version: str
    instance_group_urls: _containers.RepeatedScalarFieldContainer[str]
    status: GKENodePool.Status
    status_message: str
    autoscaling: NodePoolAutoscaling
    management: NodeManagement
    max_pods_constraint: MaxPodsConstraint
    conditions: _containers.RepeatedCompositeFieldContainer[StatusCondition]
    pod_ipv4_cidr_size: int
    upgrade_settings: GKENodePool.UpgradeSettings
    placement_policy: GKENodePool.PlacementPolicy
    update_info: GKENodePool.UpdateInfo
    etag: str
    queued_provisioning: GKENodePool.QueuedProvisioning
    best_effort_provisioning: BestEffortProvisioning
    def __init__(
        self,
        name: _Optional[str] = ...,
        config: _Optional[_Union[NodeConfig, _Mapping]] = ...,
        initial_node_count: _Optional[int] = ...,
        locations: _Optional[_Iterable[str]] = ...,
        self_link: _Optional[str] = ...,
        version: _Optional[str] = ...,
        instance_group_urls: _Optional[_Iterable[str]] = ...,
        status: _Optional[_Union[GKENodePool.Status, str]] = ...,
        status_message: _Optional[str] = ...,
        autoscaling: _Optional[_Union[NodePoolAutoscaling, _Mapping]] = ...,
        management: _Optional[_Union[NodeManagement, _Mapping]] = ...,
        max_pods_constraint: _Optional[_Union[MaxPodsConstraint, _Mapping]] = ...,
        conditions: _Optional[_Iterable[_Union[StatusCondition, _Mapping]]] = ...,
        pod_ipv4_cidr_size: _Optional[int] = ...,
        upgrade_settings: _Optional[_Union[GKENodePool.UpgradeSettings, _Mapping]] = ...,
        placement_policy: _Optional[_Union[GKENodePool.PlacementPolicy, _Mapping]] = ...,
        update_info: _Optional[_Union[GKENodePool.UpdateInfo, _Mapping]] = ...,
        etag: _Optional[str] = ...,
        queued_provisioning: _Optional[_Union[GKENodePool.QueuedProvisioning, _Mapping]] = ...,
        best_effort_provisioning: _Optional[_Union[BestEffortProvisioning, _Mapping]] = ...,
    ) -> None: ...



from __future__ import annotations

import itertools
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Iterator, TypeVar

import pyarrow as pa
import zoneinfo
from packaging.version import parse
from typing_extensions import TypeGuard

from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl

_logger = get_logger(__name__)

try:
    import orjson

    json_loads = orjson.loads
except ImportError:
    import json

    json_loads = json.loads


def is_version_gte(version: str, target: str) -> bool:
    return parse(version) >= parse(target)


try:
    import polars as pl

    is_new_polars = is_version_gte(pl.__version__, "0.18.0")
    polars_has_pad_start = is_version_gte(pl.__version__, "0.19.12")
except ImportError:
    is_new_polars = False
    polars_has_pad_start = False


def chunked_df_slices(df: pl.LazyFrame | pl.DataFrame, chunk_size: int) -> Iterator[pl.DataFrame]:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if len(df.columns) == 0:
        return
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    if len(df) == 0:
        return
    if chunk_size == -1:
        yield df
        return
    assert chunk_size > 0, "Chunk size must be -1 (for no chunking) or positive"
    for i in itertools.count():
        df_slice = df.slice(offset=i * chunk_size, length=chunk_size)
        if len(df_slice) == 0:
            # _logger.info("Chunk has length of 0; breaking")
            # No more data to write!
            break
        # _logger.info(f"Yielding chunk {i} of size {len(df_slice)}")
        yield df_slice


def pl_datetime_to_iso_string(expr: pl.Expr, tz_key: str | None) -> pl.Expr:
    """Convert a datetime expression, optionally with a timezone, into an ISO-formatted string
    The ``tz_key`` should be a timezone understood by ``zoneinfo``.
    """
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if tz_key is None:
        timezone = ""
    else:
        tzinfo = zoneinfo.ZoneInfo(tz_key)
        utc_offset = tzinfo.utcoffset(None)
        if utc_offset is None:
            raise ValueError(f"Timezone has no UTC offset: {tz_key}")
        sign = "-" if utc_offset < timedelta(0) else "+"
        seconds = abs(utc_offset.seconds + utc_offset.days * 24 * 3600)
        hours = seconds // 3600
        minutes = seconds % 3600
        timezone = f"{sign}{hours:02d}:{minutes:02d}"
    if polars_has_pad_start:
        return pl.format(
            "{}-{}-{}T{}:{}:{}.{}" + timezone,
            expr.dt.year().cast(pl.Utf8).str.pad_start(4, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.month().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.day().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.hour().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.minute().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.second().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.microsecond().cast(pl.Utf8).str.pad_start(6, "0"),  # pyright: ignore -- polars backcompat
        )
    else:
        return pl.format(
            "{}-{}-{}T{}:{}:{}.{}" + timezone,
            expr.dt.year().cast(pl.Utf8).str.rjust(4, "0"),
            expr.dt.month().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.day().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.hour().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.minute().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.second().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.microsecond().cast(pl.Utf8).str.rjust(6, "0"),
        )


def pl_date_to_iso_string(expr: pl.Expr) -> pl.Expr:
    """Convert a date expression into an ISO-formatted string"""
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if is_new_polars:
        return pl.format(
            "{}-{}-{}",
            expr.dt.year().cast(pl.Utf8).str.pad_start(4, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.month().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.day().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
        )
    else:
        return pl.format(
            "{}-{}-{}",
            expr.dt.year().cast(pl.Utf8).str.rjust(4, "0"),
            expr.dt.month().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.day().cast(pl.Utf8).str.rjust(2, "0"),
        )


def pl_time_to_iso_string(expr: pl.Expr) -> pl.Expr:
    """Convert a time expression into an ISO-formatted string"""
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if is_new_polars:
        return pl.format(
            "{}:{}:{}.{}",
            expr.dt.hour().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.minute().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.second().cast(pl.Utf8).str.pad_start(2, "0"),  # pyright: ignore -- polars backcompat
            expr.dt.microsecond().cast(pl.Utf8).str.pad_start(6, "0"),  # pyright: ignore -- polars backcompat
        )
    else:
        return pl.format(
            "{}:{}:{}.{}",
            expr.dt.hour().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.minute().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.second().cast(pl.Utf8).str.rjust(2, "0"),
            expr.dt.microsecond().cast(pl.Utf8).str.rjust(6, "0"),
        )


def pl_json_decode(series: pl.Series, dtype: pl.PolarsDataType | None = None) -> pl.Series:
    if is_new_polars:
        decoded_series = series.map_elements(json_loads, return_dtype=dtype)  # pyright: ignore -- polars backcompat
    else:
        decoded_series = series.apply(json_loads, return_dtype=dtype)
    if dtype is not None:
        # Special case -- for nested dtypes polars doesn't always respect the return_dtype
        decoded_series = decoded_series.cast(dtype)
    return decoded_series


def pl_duration_to_iso_string(expr: pl.Expr) -> pl.Expr:
    """Convert a duration expression into an ISO-formatted string"""
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    return pl.format(
        "{}P{}DT{}H{}M{}.{}S",
        pl.when(expr.dt.microseconds() < 0).then(pl.lit("-")).otherwise(pl.lit("")),
        expr.dt.days().abs().cast(pl.Utf8),
        (expr.dt.hours().abs() % 24).cast(pl.Utf8),
        (expr.dt.minutes().abs() % 60).cast(pl.Utf8),
        (expr.dt.seconds().abs() % 60).cast(pl.Utf8),
        (expr.dt.microseconds().abs() % 1_000_000)
        .cast(pl.Utf8)
        .str.pad_start(6, "0")  # pyright: ignore -- polars backcompat
        if is_new_polars
        else (expr.dt.microseconds().abs() % 1_000_000).cast(pl.Utf8).str.rjust(6, "0"),
    )


def pl_json_encode(expr: pl.Expr, dtype: pl.PolarsDataType):
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if isinstance(dtype, pl.Struct):
        # Polars does not distinguish between none and an empty struct
        return _json_encode_inner(expr, dtype)
    else:
        return pl.when(expr.is_null()).then(pl.lit("null", dtype=pl.Utf8)).otherwise(_json_encode_inner(expr, dtype))


def _py_escape_str(x: str) -> str:
    # See https://stackoverflow.com/questions/4901133/json-and-escaping-characters for the characters that must be escaped
    return (
        x.replace("\\", "\\\\")
        .replace("\u0000", "\\u0000")
        .replace("\u0001", "\\u0001")
        .replace("\u0002", "\\u0002")
        .replace("\u0003", "\\u0003")
        .replace("\u0004", "\\u0004")
        .replace("\u0005", "\\u0005")
        .replace("\u0006", "\\u0006")
        .replace("\u0007", "\\u0007")
        .replace("\b", "\\b")  # equal to \u0008
        .replace("\t", "\\t")  # equal to \u0009
        .replace("\n", "\\n")  # equal to \u000a
        .replace("\u000b", "\\u000b")
        .replace("\f", "\\f")  # equal to \u000c
        .replace("\r", "\\r")  # equal to \u000d
        .replace("\u000e", "\\u000e")
        .replace("\u000f", "\\u000f")
        .replace("\u0010", "\\u0010")
        .replace("\u0011", "\\u0011")
        .replace("\u0012", "\\u0012")
        .replace("\u0013", "\\u0013")
        .replace("\u0014", "\\u0014")
        .replace("\u0015", "\\u0015")
        .replace("\u0016", "\\u0016")
        .replace("\u0017", "\\u0017")
        .replace("\u0018", "\\u0018")
        .replace("\u0019", "\\u0019")
        .replace("\u001a", "\\u001a")
        .replace("\u001b", "\\u001b")
        .replace("\u001c", "\\u001c")
        .replace("\u001d", "\\u001d")
        .replace("\u001e", "\\u001e")
        .replace("\u001f", "\\u001f")
        .replace('"', '\\"')
    )


def pl_escape_str(x: pl.Expr) -> pl.Expr:
    # Not sure if using a regex would be faster, but literal expressions are easier to write
    # See https://stackoverflow.com/questions/4901133/json-and-escaping-characters for the characters that must be escaped
    return (
        x.str.replace_all("\\", "\\\\", literal=True)
        .str.replace_all("\u0000", "\\u0000", literal=True)
        .str.replace_all("\u0001", "\\u0001", literal=True)
        .str.replace_all("\u0002", "\\u0002", literal=True)
        .str.replace_all("\u0003", "\\u0003", literal=True)
        .str.replace_all("\u0004", "\\u0004", literal=True)
        .str.replace_all("\u0005", "\\u0005", literal=True)
        .str.replace_all("\u0006", "\\u0006", literal=True)
        .str.replace_all("\u0007", "\\u0007", literal=True)
        .str.replace_all("\b", "\\b", literal=True)  # equal to \u0008
        .str.replace_all("\t", "\\t", literal=True)  # equal to \u0009
        .str.replace_all("\n", "\\n", literal=True)  # equal to \u000a
        .str.replace_all("\u000b", "\\u000b", literal=True)
        .str.replace_all("\f", "\\f", literal=True)  # equal to \u000c
        .str.replace_all("\r", "\\r", literal=True)  # equal to \u000d
        .str.replace_all("\u000e", "\\u000e", literal=True)
        .str.replace_all("\u000f", "\\u000f", literal=True)
        .str.replace_all("\u0010", "\\u0010", literal=True)
        .str.replace_all("\u0011", "\\u0011", literal=True)
        .str.replace_all("\u0012", "\\u0012", literal=True)
        .str.replace_all("\u0013", "\\u0013", literal=True)
        .str.replace_all("\u0014", "\\u0014", literal=True)
        .str.replace_all("\u0015", "\\u0015", literal=True)
        .str.replace_all("\u0016", "\\u0016", literal=True)
        .str.replace_all("\u0017", "\\u0017", literal=True)
        .str.replace_all("\u0018", "\\u0018", literal=True)
        .str.replace_all("\u0019", "\\u0019", literal=True)
        .str.replace_all("\u001a", "\\u001a", literal=True)
        .str.replace_all("\u001b", "\\u001b", literal=True)
        .str.replace_all("\u001c", "\\u001c", literal=True)
        .str.replace_all("\u001d", "\\u001d", literal=True)
        .str.replace_all("\u001e", "\\u001e", literal=True)
        .str.replace_all("\u001f", "\\u001f", literal=True)
        .str.replace_all('"', '\\"', literal=True)
    )


def _backup_json_encode(x: Any) -> str:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if isinstance(x, pl.Series):
        x = x.to_list()
    return orjson.dumps(x, option=orjson.OPT_SORT_KEYS | orjson.OPT_UTC_Z | orjson.OPT_NAIVE_UTC).decode("utf8")


T = TypeVar("T", bound=type)


def _check_is_type(dtype: pl.PolarsDataType, typ: T) -> TypeGuard[T]:
    # polars < 0.20
    if isinstance(dtype, type):
        return issubclass(dtype, typ)
    else:
        # polars >= 0.20
        return isinstance(dtype, typ)


def _json_encode_inner(expr: pl.Expr, dtype: pl.PolarsDataType) -> pl.Expr:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if isinstance(dtype, (type, pl.DataType)):  # pyright: ignore [reportUnnecessaryIsInstance]
        if _check_is_type(dtype, pl.Boolean):
            return pl.when(expr).then(pl.lit("true", dtype=pl.Utf8)).otherwise(pl.lit("false", dtype=pl.Utf8))
        if _check_is_type(dtype, (pl.Float32, pl.Float64)):  # pyright: ignore[reportArgumentType]
            # Floats of nan, +inf, or -inf cannot be represented as json, so instead convert them to "null"
            # Otherwise, they can be cast directly
            return (
                pl.when(expr.is_nan() | expr.is_infinite())
                .then(pl.lit("null", dtype=pl.Utf8))
                .otherwise(expr.cast(pl.Utf8))
            )

        if _check_is_type(
            dtype,
            (
                pl.Float32,
                pl.Float64,
                pl.Int16,
                pl.Int32,
                pl.Int64,
                pl.Int8,
                pl.UInt16,
                pl.UInt32,
                pl.UInt64,
                pl.UInt8,
            ),  # pyright: ignore[reportArgumentType]
        ):
            # Ints can be cast directly
            return expr.cast(pl.Utf8, strict=True)
        if _check_is_type(dtype, pl.Utf8):
            # Must escape any quote and backslashes
            # First escape all backslashes to double backslash
            # Then, wrap the result in quotes

            return pl.format('"{}"', pl_escape_str(expr))

        if _check_is_type(dtype, pl.Time):
            # Convert time to iso
            return pl_json_encode(pl_time_to_iso_string(expr), pl.Utf8)
        if _check_is_type(dtype, pl.Date):
            # Convert a date to iso
            return pl_json_encode(pl_date_to_iso_string(expr), pl.Utf8)
    if isinstance(dtype, pl.Datetime):
        # Convert datetime to iso
        return pl_json_encode(pl_datetime_to_iso_string(expr, dtype.time_zone), pl.Utf8)
    if isinstance(dtype, pl.Duration):
        # Convert duration to iso
        return pl_json_encode(pl_duration_to_iso_string(expr), pl.Utf8)
    if isinstance(dtype, pl.List):
        inner_dtype = dtype.inner
        assert inner_dtype is not None
        # TODO -- this will likely break on lists of structs or lists of lists
        # However, .eval cannot be called on an empty list
        # So we append an empty element, eval, and then remove it
        if isinstance(inner_dtype, (pl.List, pl.Struct)):
            # TODO: We do NOT support nested collections right now, because of how we must append a default
            # value in the case of an empty collection
            # Need to add support for this eventually
            # For now, use a (slow) UDF
            _logger.warning(f"Including the python UDF json encode in the polars expression to handle dtype {dtype}")
            if is_new_polars:
                return expr.map_elements(  # pyright: ignore -- polars backcompat
                    _backup_json_encode, return_dtype=pl.Utf8
                )
            else:
                return expr.apply(_backup_json_encode, return_dtype=pl.Utf8)
        expr = expr.fill_null([])
        lists_with_extra_none = (
            expr.list.concat(pl.lit(None))  # pyright: ignore -- back compat
            if is_new_polars
            else expr.arr.concat(pl.lit(None))  # pyright: ignore -- back compat
        )
        encoded_with_extra = (
            lists_with_extra_none.list.eval(  # pyright: ignore -- polars backcompat
                pl_json_encode(pl.element(), inner_dtype)
            )
            if is_new_polars
            else lists_with_extra_none.arr.eval(  # pyright: ignore -- back compat
                pl_json_encode(pl.element(), inner_dtype)
            )
        )
        encoded_without_extra = (
            encoded_with_extra.list.slice(offset=0, length=expr.list.len())  # pyright: ignore -- polars backcompat
            if is_new_polars
            else encoded_with_extra.arr.slice(offset=0, length=expr.arr.lengths())  # pyright: ignore -- back compat
        )
        return (
            pl.format("[{}]", encoded_without_extra.list.join(","))  # pyright: ignore -- polars backcompat
            if is_new_polars
            else pl.format("[{}]", encoded_without_extra.arr.join(","))  # pyright: ignore -- back compat
        )
    if isinstance(dtype, pl.Struct):
        fields_encoded = [
            pl.format((f'"{_py_escape_str(f.name)}":' "{}"), pl_json_encode(expr.struct.field(f.name), f.dtype))
            for f in sorted(dtype.fields, key=lambda x: x.name)
        ]
        format_str = "{" + ",".join(["{}"] * len(fields_encoded)) + "}"
        if len(fields_encoded) == 0:
            return pl.lit("{}")
        return pl.format(format_str, *fields_encoded)

    if dtype == pl.Binary:  # pyright: ignore[reportUnnecessaryComparison]
        return pl_json_encode(expr.bin.encode("base64"), pl.Utf8)

    raise TypeError(f"Unsupported dtype for json encoding: {dtype}")


def recursively_has_float16(dtype: pa.DataType) -> bool:
    """Check whether this dtype has a float16 field in it"""
    if pa.types.is_float16(dtype):
        return True
    if pa.types.is_struct(dtype):
        assert isinstance(dtype, pa.StructType)
        return any(recursively_has_float16(dtype.field(i).type) for i in range(dtype.num_fields))
    if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
        assert isinstance(dtype, (pa.LargeListType, pa.ListType, pa.FixedSizeListType))
        return recursively_has_float16(dtype.value_type)
    if pa.types.is_map(dtype):
        assert isinstance(dtype, pa.MapType)
        return recursively_has_float16(dtype.key_type) or recursively_has_float16(dtype.item_type)
    return False


def pl_is_uniquable_on(dtype: pl.PolarsDataType) -> bool:
    """Check whether the Polars dtype can be uniqued upon, which currently tests for existence of lists in the dtype"""
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if isinstance(dtype, pl.Struct):
        return all(pl_is_uniquable_on(f.dtype) for f in dtype.fields)
    if isinstance(dtype, pl.List):
        return False
    return True


def pa_is_uniquable_on(dtype: pa.DataType) -> bool:
    """Check whether the PyArrow dtype can be uniqued upon, which currently tests for existence of lists in the dtype"""
    if pa.types.is_struct(dtype):
        assert isinstance(dtype, pa.StructType)
        return all(pa_is_uniquable_on(dtype.field(i).type) for i in range(dtype.num_fields))
    if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
        return False
    return True


def recursively_has_struct(dtype: pa.DataType) -> bool:
    """Check whether this dtype has a struct field in it"""
    if pa.types.is_struct(dtype):
        return True
    if pa.types.is_list(dtype) or pa.types.is_large_list(dtype) or pa.types.is_fixed_size_list(dtype):
        assert isinstance(dtype, (pa.LargeListType, pa.ListType, pa.FixedSizeListType))
        return recursively_has_struct(dtype.value_type)
    if pa.types.is_map(dtype):
        assert isinstance(dtype, pa.MapType)
        return recursively_has_struct(dtype.key_type) or recursively_has_struct(dtype.item_type)
    return False



from __future__ import annotations

from datetime import datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    List,
    Literal,
    Mapping,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import TypedDict

from chalk._validation.feature_validation import FeatureValidation
from chalk._validation.validation import Validation
from chalk.features._encoding.primitive import TPrimitive
from chalk.features.underscore import Underscore
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import CHALK_MAX_TIMEDELTA, CronTab, Duration, parse_chalk_duration_s

TPrim = TypeVar("TPrim", bound=TPrimitive)
TRich = TypeVar("TRich")

if TYPE_CHECKING:
    import pyarrow as pa

    from chalk.features._encoding.converter import TDecoder, TEncoder
    from chalk.features.feature_field import WindowConfigResolved


class WindowedInstance(Generic[TRich]):
    def __init__(self, values: Mapping[str, TRich]):
        super().__init__()
        self.values = values

    def __call__(self, period: str):
        return self.values[period]


class WindowedMeta(type, Generic[TRich]):
    def __getitem__(cls, underlying: Type[Any]) -> Windowed[TRich]:
        return Windowed(
            kind=underlying,
            buckets=[],
            description=None,
            owner=None,
            tags=None,
            name=None,
            default=...,
            max_staleness=None,
            version=None,
            etl_offline_to_online=None,
            encoder=None,
            decoder=None,
            min=None,
            max=None,
            min_length=None,
            max_length=None,
            contains=None,
            strict=False,
            dtype=None,
            validations=None,
            offline_ttl=None,
            expression=None,
            materialization=None,
        )  # noqa


JsonValue = Any


def get_name_with_duration(name_or_fqn: str, duration: Union[str, int, timedelta]) -> str:
    duration_secs = parse_chalk_duration_s(duration)
    if duration_secs >= CHALK_MAX_TIMEDELTA.total_seconds():
        return f"{name_or_fqn}__all__"
    return f"{name_or_fqn}__{duration_secs}__"


if TYPE_CHECKING:
    _WINDOWED_METACLASS = type
else:
    _WINDOWED_METACLASS = WindowedMeta


class Windowed(Generic[TRich], metaclass=_WINDOWED_METACLASS):
    """Declare a windowed feature.

    Examples
    --------
    >>> @features
    ... class User:
    ...     failed_logins: Windowed[int] = windowed("10m", "24h")
    """

    def __getitem__(self, item: Type[Any] | Any):
        # Here for editor support
        super().__getitem__(item)  # pyright: ignore

    @property
    def buckets_seconds(self) -> Set[int]:
        return {parse_chalk_duration_s(bucket) for bucket in self._buckets}

    @property
    def kind(self) -> Type[TRich]:
        if self._kind is None:
            raise RuntimeError("Window type has not yet been parsed")
        return self._kind

    @kind.setter
    def kind(self, kind: Type[TRich]) -> None:
        assert self._kind is None, "Window type cannot be set twice"
        self._kind = kind

    def _to_feature(self, bucket: int | str | None):
        from chalk.features import Feature

        assert self._name is not None

        window_duration = None if bucket is None else parse_chalk_duration_s(bucket)
        if bucket is None:
            name = self._name
        elif window_duration not in self.buckets_seconds:
            raise ValueError(f"Bucket {bucket} is not in the list of specified buckets")
        else:
            name = get_name_with_duration(self._name, duration=window_duration)

        return Feature(
            name=name,
            version=self._version,
            owner=self._owner,
            tags=None if self._tags is None else list(ensure_tuple(self._tags)),
            description=self._description,
            primary=False,
            default=self._default,
            max_staleness=self._max_staleness,
            offline_ttl=self._offline_ttl,
            etl_offline_to_online=self._etl_offline_to_online,
            encoder=self._encoder,
            decoder=self._decoder,
            pyarrow_dtype=self._dtype,
            validations=FeatureValidation(
                min=self._min,
                max=self._max,
                min_length=self._min_length,
                max_length=self._max_length,
                contains=self._contains,
                strict=self._strict,
            ),
            all_validations=(
                None
                if self._validations is None
                else [
                    FeatureValidation(
                        min=v.min,
                        max=v.max,
                        min_length=v.min_length,
                        max_length=v.max_length,
                        contains=None,
                        strict=v.strict,
                    )
                    for v in self._validations
                ]
            ),
            # Only the root feature should have all the durations
            # The pseudo-features, which are bound to a duration, should not have the durations
            # of the other buckets
            window_durations=tuple(self.buckets_seconds) if bucket is None else tuple(),
            window_duration=window_duration,
            underscore_expression=self._expression,
            window_materialization=(
                MaterializationWindowConfig(bucket_duration=timedelta(seconds=window_duration))
                if self._materialization is True and window_duration is not None
                else self._materialization
                if isinstance(self._materialization, dict)
                else None
            ),
        )

    def __init__(
        self,
        buckets: List[str],
        description: str | None,
        owner: str | None,
        tags: Any | None,
        name: str | None,
        default: TRich | ellipsis,
        max_staleness: Duration | ellipsis | None,
        version: int | None,
        etl_offline_to_online: bool | None,
        encoder: TEncoder[TPrim, TRich] | None,
        decoder: TDecoder[TPrim, TRich] | None,
        min: TRich | None,
        max: TRich | None,
        min_length: int | None,
        max_length: int | None,
        contains: TRich | None,
        strict: bool,
        validations: List[Validation] | None,
        dtype: pa.DataType | None,
        kind: Type[TRich] | None,
        offline_ttl: Duration | ellipsis | None,
        expression: Underscore | None,
        materialization: MaterializationWindowConfig | Literal[True] | None,
    ):
        super().__init__()
        self._kind = kind
        self._buckets = buckets
        self._description = description
        self._owner = owner
        self._tags = tags
        self._name = name
        self._default = default
        self._max_staleness = max_staleness
        self._offline_ttl = offline_ttl
        self._description = description
        self._version = version
        self._etl_offline_to_online = etl_offline_to_online
        self._encoder = encoder
        self._decoder = decoder
        self._min = min
        self._max = max
        self._min_length = min_length
        self._max_length = max_length
        self._contains = contains
        self._strict = strict
        self._validations = validations
        self._dtype = dtype
        self._expression = expression
        self._materialization = materialization


class SelectedWindow:
    def __init__(self, kind: Windowed, selected: str):
        super().__init__()
        self.windowed = kind
        self.selected = selected


class MaterializationWindowConfig(TypedDict, total=False):
    """Configuration for window aggregates. At least one of
    `bucket_duration` and `bucket_durations` must be provided.

    If both are provided, `bucket_duration` acts as a default
    for the window materialization, which may be overridden by
    `bucket_duration`.
    """

    bucket_duration: Duration
    """The duration of each bucket in the window, using a
    `chalk.Duration` string, e.g. `"1m"`, `"1h"`, `"1d"`.

    To use different bucket durations for different window
    sizes, see `bucket_durations` below.
    """

    bucket_durations: Mapping[Duration, Sequence[Duration] | Duration]
    """A mapping from the desired bucket duration to the
    window size(s) that should use that bucket duration.

    If `bucket_duration` is also provided, any window
    durations not specified in this mapping will pick up
    the bucket duration from the `bucket_duration` parameter.

    This parameter is useful when you have some very large
    windows and some very small windows. For example, if you
    have a 365-day window and a 10-minute window, you wouldn't
    want to maintain 365 days of 10-minute buckets in the online
    store. However, using a 1-day bucket for the 10-minute
    window would also lead to significantly more events fitting
    into the window than you might want.

    In this case, you could specify:

    ```
    count: Windowed[int] = windowed(
        "1d", "7d", "60d", "365d",
        materialization={
            # 1-day buckets as a default, for the 7d window
            bucket_duration="1d",
            bucket_durations={
                # 10-minute buckets for the 1d window
                "10m": "1d",
                # 5-day buckets for 60d and 365d windows
                "5d": ["60d", "365d"],
            }
        },
        expression=_.events.count(),
    )
    ```
    """

    bucket_start: datetime
    """The lower bound of the first bucket. All buckets will start at some multiple of the bucket duration after this time."""

    bucket_starts: Mapping[datetime, Sequence[Duration] | Duration]
    """Used to specify a different bucket start for each window duration. Same format as `bucket_durations`."""

    backfill_schedule: CronTab | None
    """The schedule on which to automatically backfill the aggregation. For example, `"* * * * *"` or `"1h"`."""

    continuous_buffer_duration: Duration | None
    """The minimum period of time for which to sample data directly via online query, rather than from the backfilled aggregations."""


def group_by_windowed(
    *buckets: str,
    expression: Underscore,
    materialization: MaterializationWindowConfig | Literal[True],
    days: Iterable[int] = (),
    hours: Iterable[int] = (),
    minutes: Iterable[int] = (),
    description: str | None = None,
    owner: str | None = None,
    tags: Any | None = None,
    name: str | None = None,
    default: TRich | ellipsis = ...,
    min: TRich | None = None,
    max: TRich | None = None,
    strict: bool = False,
    validations: List[Validation] | None = None,
    dtype: pa.DataType | None = None,
) -> Any:
    """Create a windowed feature with grouping.

    See more at https://docs.chalk.ai/docs/materialized_aggregations

    Parameters
    ----------
    buckets
        The size of the buckets for the window function.
        Buckets are specified as strings in the format `"1d"`, `"2h"`, `"1h30m"`, etc.
        You may also choose to specify the buckets using the days, hours, and minutes
        parameters instead. The buckets parameter is helpful if you want to use
        multiple units to express the bucket size, like `"1h30m"`.
    days
        Convenience parameter for specifying the buckets in days.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1d"`.
    hours
        Convenience parameter for specifying the buckets in hours.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1h"`.
    minutes
        Convenience parameter for specifying the buckets in minutes.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1m"`.
    default
        The default value of the feature if it otherwise can't be computed.
    materialization
        Configuration for aggregating data. Pass `bucket_duration` with a
        [Duration](/api-docs#Duration) to configure the bucket size for aggregation.

        See more at https://docs.chalk.ai/docs/materialized_aggregations
    dtype
        The backing `pyarrow.DataType` for the feature. This parameter can
        be used to control the storage format of data. For example, if you
        have a lot of data that could be represented as smaller data types,
        you can use this parameter to save space.

        >>> import pyarrow as pa
        >>> from chalk.features import features
        >>> from chalk.streams import Windowed, windowed
        >>> @features
        ... class User:
        ...     id: str
        ...     email_count: Windowed[int] = windowed(
        ...         "10m", "30m",
        ...         dtype=pa.int16(),
        ...     )
    owner
        You may also specify which person or group is responsible for a feature.
        The owner tag will be available in Chalk's web portal.
        Alerts that do not otherwise have an owner will be assigned
        to the owner of the monitored feature.
    tags
        Add metadata to a feature for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
    min
        If specified, when this feature is computed, Chalk will check that `x >= min`.
    max
        If specified, when this feature is computed, Chalk will check that `x <= max`.
    strict
        If `True`, if this feature does not meet the validation criteria, Chalk will not persist
        the feature value and will treat it as failed.
    expression
        The expression to compute the feature. This is an underscore expression,
        like `_.transactions[_.amount].sum()`.

    Other Parameters
    ----------------
    name
        The name for the feature. By default, the name of a feature is
        the name of the attribute on the class, prefixed with
        the camel-cased name of the class. Note that if you provide an
        explicit name, the namespace, determined by the feature class,
        will still be prepended. See `features` for more details.
    description
        Descriptions are typically provided as comments preceding
        the feature definition. For example, you can document a
        `emails_by_category` feature with information about the values
        as follows:
        >>> from chalk import _
        >>> @features
        ... class User:
        ...     # Count of emails sent, grouped by category
        ...     emails_by_category: DataFrame = group_by_windowed(
        ...         "10m", "30m",
        ...         description="Count of emails sent, grouped by category",
        ...         expression=_.emails.group_by(_.category).count(),
        ...     )

        You can also specify the description directly with this parameter:
        >>> from chalk import _
        >>> @features
        ... class User:
        ...     emails: DataFrame[Email]
        ...     emails_by_category: DataFrame = group_by_windowed(
        ...         "10m", "30m",
        ...         description="Count of emails sent, grouped by category",
        ...         expression=_.emails.group_by(_.category).count(),
        ...     )
    validations

    Returns
    -------
    GroupByWindowed

    Examples
    --------
    >>> from chalk import group_by_windowed, DataFrame
    >>> from chalk.features import features
    >>> @features
    ... class Email:
    ...     id: int
    ...     user_id: "User.id"
    >>> @features
    ... class User:
    ...     id: int
    ...     emails: DataFrame[Email]
    ...     emails_by_category: DataFrame = group_by_windowed(
    ...         "10m", "30m",
    ...         expression=_.emails.group_by(_.category).count(),
    ...     )
    """
    return GroupByWindowed(
        list(buckets) + [f"{x}m" for x in minutes] + [f"{x}h" for x in hours] + [f"{x}d" for x in days],
        description=description,
        owner=owner,
        tags=tags,
        name=name,
        default=default,
        min=min,
        max=max,
        strict=strict,
        dtype=dtype,
        validations=validations,
        expression=expression,
        materialization=materialization,
    )


class GroupByWindowNarrowed(Generic[TRich]):
    def __init__(
        self,
        kind: GroupByWindowed[TRich],
        filters: Tuple[Any, ...],
        selected_window: int,
    ):
        super().__init__()
        self._kind = kind
        self._filters = filters
        self._selected_window = selected_window

    def __str__(self):
        parent_name = self._kind._name  # pyright: ignore[reportPrivateUsage]
        namespace = self._kind._namespace  # pyright: ignore[reportPrivateUsage]
        return f"{namespace}.{parent_name}_{self._selected_window}"


class GroupByWindowed(Generic[TRich], metaclass=_WINDOWED_METACLASS):
    def __init__(
        self,
        buckets: List[str],
        materialization: MaterializationWindowConfig | Literal[True],
        expression: Underscore,
        description: str | None,
        owner: str | None,
        tags: Any | None,
        name: str | None,
        default: Union[TRich, ellipsis],
        min: TRich | None,
        max: TRich | None,
        strict: bool,
        validations: List[Validation] | None,
        dtype: pa.DataType | None,
    ):
        super().__init__()
        self._buckets = buckets
        self._description = description
        self._owner = owner
        self._tags = tags
        self._name = name
        self._namespace: str | None = None
        self._default = default
        self._description = description
        self._min = min
        self._max = max
        self._strict = strict
        self._validations = validations
        self._dtype = dtype
        self._expression = expression
        self._materialization: MaterializationWindowConfig | Literal[True] = materialization
        self._kind: Type | None = None
        self._window_materialization_parsed: "WindowConfigResolved | None" = None

    def __getitem__(self, item: Type[Any]):
        filters = []
        window_seconds = []
        for x in ensure_tuple(item):
            if isinstance(x, str):
                duration_seconds = parse_chalk_duration_s(x)
                if duration_seconds not in self.buckets_seconds:
                    raise ValueError(f"Bucket {x} is not in the list of specified buckets")
                window_seconds.append(duration_seconds)
            else:
                filters.append(x)

        if len(window_seconds) == 0:
            raise ValueError("No window selected")
        elif len(window_seconds) > 1:
            raise ValueError("Only one window can be selected")

        return GroupByWindowNarrowed(
            kind=self,
            filters=tuple(filters),
            selected_window=window_seconds[0],
        )

    @property
    def buckets_seconds(self) -> Set[int]:
        return {parse_chalk_duration_s(bucket) for bucket in self._buckets}


def windowed(
    *buckets: str,
    days: Iterable[int] = (),
    hours: Iterable[int] = (),
    minutes: Iterable[int] = (),
    description: str | None = None,
    owner: str | None = None,
    tags: Any | None = None,
    name: str | None = None,
    default: TRich | ellipsis = ...,
    max_staleness: Duration | ellipsis | None = ...,
    offline_ttl: Duration | ellipsis | None = ...,
    version: int | None = None,
    etl_offline_to_online: bool | None = None,
    encoder: TEncoder[TPrim, TRich] | None = None,
    decoder: TDecoder[TPrim, TRich] | None = None,
    min: TRich | None = None,
    max: TRich | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    strict: bool = False,
    validations: List[Validation] | None = None,
    dtype: pa.DataType | None = None,
    expression: Underscore | None = None,
    materialization: MaterializationWindowConfig | Literal[True] | None = None,
) -> Windowed[TRich]:
    """Create a windowed feature.

    See more at https://docs.chalk.ai/docs/aggregations

    Parameters
    ----------
    buckets
        The size of the buckets for the window function.
        Buckets are specified as strings in the format `"1d"`, `"2h"`, `"1h30m"`, etc.
        You may also choose to specify the buckets using the days, hours, and minutes
        parameters instead. The buckets parameter is helpful if you want to use
        multiple units to express the bucket size, like `"1h30m"`.
    days
        Convenience parameter for specifying the buckets in days.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1d"`.
    hours
        Convenience parameter for specifying the buckets in hours.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1h"`.
    minutes
        Convenience parameter for specifying the buckets in minutes.
        Using this parameter is equvalent to specifying the buckets parameter
        with a string like `"1m"`.
    default
        The default value of the feature if it otherwise can't be computed.
    materialization
        Configuration for aggregating data. Pass `bucket_duration` with a
        [Duration](/api-docs#Duration) to configure the bucket size for aggregation.
        If `True`, each of the windows will use a bucket duration equal to its
        window duration.

        See more at https://docs.chalk.ai/docs/materialized_aggregations
    owner
        You may also specify which person or group is responsible for a feature.
        The owner tag will be available in Chalk's web portal.
        Alerts that do not otherwise have an owner will be assigned
        to the owner of the monitored feature.
    tags
        Add metadata to a feature for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
    max_staleness
        When a feature is expensive or slow to compute, you may wish to cache its value.
        Chalk uses the terminology "maximum staleness" to describe how recently a feature
        value needs to have been computed to be returned without re-running a resolver.

        See more at https://docs.chalk.ai/docs/feature-caching
    version
        Feature versions allow you to manage a feature as its
        definition changes over time.

        The `version` keyword argument allows you to specify the
        maximum number of versions available for this feature.

        See more at https://docs.chalk.ai/docs/feature-versions
    etl_offline_to_online
        When `True`, Chalk copies this feature into the online environment
        when it is computed in offline resolvers.

        See more at https://docs.chalk.ai/docs/reverse-etl
    min
        If specified, when this feature is computed, Chalk will check that `x >= min`.
    max
        If specified, when this feature is computed, Chalk will check that `x <= max`.
    min_length
        If specified, when this feature is computed, Chalk will check that `len(x) >= min_length`.
    max_length
        If specified, when this feature is computed, Chalk will check that `len(x) <= max_length`.
    strict
        If `True`, if this feature does not meet the validation criteria, Chalk will not persist
        the feature value and will treat it as failed.
    expression
        The expression to compute the feature. This is an underscore expression, like `_.transactions[_.amount].sum()`.
    validations
        A list of Validations to apply to this feature.

        See more at https://docs.chalk.ai/api-docs#Validation
    offline_ttl
        Sets a maximum age for values eligible to be retrieved from the offline store,
        defined in relation to the query's current point-in-time.

    Other Parameters
    ----------------
    name
        The name for the feature. By default, the name of a feature is
        the name of the attribute on the class, prefixed with
        the camel-cased name of the class. Note that if you provide an
        explicit name, the namespace, determined by the feature class,
        will still be prepended. See `features` for more details.
    description
        Descriptions are typically provided as comments preceding
        the feature definition. For example, you can document a
        `email_count` feature with information about the values
        as follows:
        >>> from chalk.features import features
        >>> from chalk.streams import Windowed, windowed
        >>> @features
        ... class User:
        ...     # Count of emails sent
        ...     email_count: Windowed[int] = windowed("10m", "30m")

        You can also specify the description directly with this parameter:
        >>> from chalk.features import features
        >>> from chalk.streams import Windowed, windowed
        >>> @features
        ... class User:
        ...     email_count: Windowed[int] = windowed(
        ...         "10m",
        ...         "30m",
        ...         description="Count of emails sent",
        ...     )
    dtype
        The backing `pyarrow.DataType` for the feature. This parameter can
        be used to control the storage format of data. For example, if you
        have a lot of data that could be represented as smaller data types,
        you can use this parameter to save space.

        >>> import pyarrow as pa
        >>> from chalk.features import features
        >>> from chalk.streams import Windowed, windowed
        >>> @features
        ... class User:
        ...     id: str
        ...     email_count: Windowed[int] = windowed(
        ...         "10m", "30m",
        ...         dtype=pa.int16(),
        ...     )
    encoder
    decoder

    Returns
    -------
    Windowed[TPrim, TRich]
        Metadata for the windowed feature, parameterized by
        `TPrim` (the primitive type of the feature) and
        `TRich` (the decoded type of the feature, if `decoder` is provided).

    Examples
    --------
    >>> from chalk.features import features
    >>> from chalk.streams import windowed, Windowed
    >>> @features
    ... class User:
    ...     id: int
    ...     email_count: Windowed[int] = windowed(days=range(1, 30))
    ...     logins: Windowed[int] = windowed("10m", "1d", "30d")
    >>> User.email_count["7d"]
    """
    return Windowed(
        list(buckets) + [f"{x}m" for x in minutes] + [f"{x}h" for x in hours] + [f"{x}d" for x in days],
        description=description,
        owner=owner,
        tags=tags,
        name=name,
        default=default,
        max_staleness=max_staleness,
        version=version,
        etl_offline_to_online=etl_offline_to_online,
        encoder=cast("TEncoder", encoder),
        decoder=decoder,
        min=min,
        max=max,
        min_length=min_length,
        max_length=max_length,
        contains=None,
        strict=strict,
        dtype=dtype,
        kind=None,
        validations=validations,
        offline_ttl=offline_ttl,
        expression=expression,
        materialization=materialization,
    )



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class GetClusterEnvironmentsRequest(_message.Message):
    __slots__ = ("cluster_name",)
    CLUSTER_NAME_FIELD_NUMBER: _ClassVar[int]
    cluster_name: str
    def __init__(self, cluster_name: _Optional[str] = ...) -> None: ...

class GetClusterEnvironmentsResponse(_message.Message):
    __slots__ = ("environment_ids",)
    ENVIRONMENT_IDS_FIELD_NUMBER: _ClassVar[int]
    environment_ids: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, environment_ids: _Optional[_Iterable[str]] = ...) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.deploy_pb2 import (
    CreateBranchFromSourceDeploymentRequest,
    CreateBranchFromSourceDeploymentResponse,
    DeployBranchRequest,
    DeployBranchResponse,
    GetActiveDeploymentsRequest,
    GetActiveDeploymentsResponse,
    GetDeploymentRequest,
    GetDeploymentResponse,
    ListDeploymentsRequest,
    ListDeploymentsResponse,
    ScaleDeploymentRequest,
    ScaleDeploymentResponse,
    SuspendDeploymentRequest,
    SuspendDeploymentResponse,
    TagDeploymentRequest,
    TagDeploymentResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class DeployServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    DeployBranch: UnaryUnaryMultiCallable[
        DeployBranchRequest,
        DeployBranchResponse,
    ]
    CreateBranchFromSourceDeployment: UnaryUnaryMultiCallable[
        CreateBranchFromSourceDeploymentRequest,
        CreateBranchFromSourceDeploymentResponse,
    ]
    GetDeployment: UnaryUnaryMultiCallable[
        GetDeploymentRequest,
        GetDeploymentResponse,
    ]
    ListDeployments: UnaryUnaryMultiCallable[
        ListDeploymentsRequest,
        ListDeploymentsResponse,
    ]
    GetActiveDeployments: UnaryUnaryMultiCallable[
        GetActiveDeploymentsRequest,
        GetActiveDeploymentsResponse,
    ]
    SuspendDeployment: UnaryUnaryMultiCallable[
        SuspendDeploymentRequest,
        SuspendDeploymentResponse,
    ]
    ScaleDeployment: UnaryUnaryMultiCallable[
        ScaleDeploymentRequest,
        ScaleDeploymentResponse,
    ]
    TagDeployment: UnaryUnaryMultiCallable[
        TagDeploymentRequest,
        TagDeploymentResponse,
    ]

class DeployServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def DeployBranch(
        self,
        request: DeployBranchRequest,
        context: ServicerContext,
    ) -> DeployBranchResponse: ...
    @abstractmethod
    def CreateBranchFromSourceDeployment(
        self,
        request: CreateBranchFromSourceDeploymentRequest,
        context: ServicerContext,
    ) -> CreateBranchFromSourceDeploymentResponse: ...
    @abstractmethod
    def GetDeployment(
        self,
        request: GetDeploymentRequest,
        context: ServicerContext,
    ) -> GetDeploymentResponse: ...
    @abstractmethod
    def ListDeployments(
        self,
        request: ListDeploymentsRequest,
        context: ServicerContext,
    ) -> ListDeploymentsResponse: ...
    @abstractmethod
    def GetActiveDeployments(
        self,
        request: GetActiveDeploymentsRequest,
        context: ServicerContext,
    ) -> GetActiveDeploymentsResponse: ...
    @abstractmethod
    def SuspendDeployment(
        self,
        request: SuspendDeploymentRequest,
        context: ServicerContext,
    ) -> SuspendDeploymentResponse: ...
    @abstractmethod
    def ScaleDeployment(
        self,
        request: ScaleDeploymentRequest,
        context: ServicerContext,
    ) -> ScaleDeploymentResponse: ...
    @abstractmethod
    def TagDeployment(
        self,
        request: TagDeploymentRequest,
        context: ServicerContext,
    ) -> TagDeploymentResponse: ...

def add_DeployServiceServicer_to_server(servicer: DeployServiceServicer, server: Server) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.datasets_pb2 import (
    GetDatasetRequest,
    GetDatasetResponse,
    GetDatasetRevisionRequest,
    GetDatasetRevisionResponse,
    ListDatasetRevisionsRequest,
    ListDatasetRevisionsResponse,
    ListDatasetsRequest,
    ListDatasetsResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class DatasetMetadataServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ListDatasets: UnaryUnaryMultiCallable[
        ListDatasetsRequest,
        ListDatasetsResponse,
    ]
    GetDataset: UnaryUnaryMultiCallable[
        GetDatasetRequest,
        GetDatasetResponse,
    ]
    ListDatasetRevisions: UnaryUnaryMultiCallable[
        ListDatasetRevisionsRequest,
        ListDatasetRevisionsResponse,
    ]
    GetDatasetRevision: UnaryUnaryMultiCallable[
        GetDatasetRevisionRequest,
        GetDatasetRevisionResponse,
    ]

class DatasetMetadataServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ListDatasets(
        self,
        request: ListDatasetsRequest,
        context: ServicerContext,
    ) -> ListDatasetsResponse: ...
    @abstractmethod
    def GetDataset(
        self,
        request: GetDatasetRequest,
        context: ServicerContext,
    ) -> GetDatasetResponse: ...
    @abstractmethod
    def ListDatasetRevisions(
        self,
        request: ListDatasetRevisionsRequest,
        context: ServicerContext,
    ) -> ListDatasetRevisionsResponse: ...
    @abstractmethod
    def GetDatasetRevision(
        self,
        request: GetDatasetRevisionRequest,
        context: ServicerContext,
    ) -> GetDatasetRevisionResponse: ...

def add_DatasetMetadataServiceServicer_to_server(servicer: DatasetMetadataServiceServicer, server: Server) -> None: ...
